# Unit test fit ForecasterEquivalentDate
# ==============================================================================
import re
import pytest
from pytest import approx
import numpy as np
import pandas as pd
from pandas.tseries.offsets import DateOffset
from skforecast.recursive import ForecasterEquivalentDate

# Fixtures
from .fixtures_forecaster_equivalent_date import y


def test_fit_TypeError_when_y_is_not_a_Series():
    """
    Test TypeError is raised when y is not a pandas Series.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    not_valid_y = np.arange(10)

    err_msg = re.escape(
        f"`y` must be a pandas Series with a DatetimeIndex or a RangeIndex. "
        f"Found {type(not_valid_y)}."
    )
    with pytest.raises(TypeError, match=err_msg):
        forecaster.fit(y=not_valid_y)


@pytest.mark.parametrize("y", 
                         [pd.Series(np.arange(10)), 
                          pd.Series(np.arange(10), index=pd.date_range(start='1/1/2021', periods=10))])
def test_fit_TypeError_offset_DateOffset_y_index_not_DatetimeIndex(y):
    """
    Test TypeError is raised when offset is a DateOffset and y index is 
    not a DatetimeIndex or has no freq.
    """
    forecaster = ForecasterEquivalentDate(
        offset=DateOffset(days=1), n_offsets=2, agg_func=np.mean
    )

    if isinstance(y.index, pd.DatetimeIndex):
        y.index.freq = None

    err_msg = re.escape(
        "If `offset` is a pandas DateOffset, the index of `y` must be a "
        "pandas DatetimeIndex with frequency."
    )
    with pytest.raises(TypeError, match=err_msg):
        forecaster.fit(y=y)


def test_fit_ValueError_length_y_less_than_window_size_offset_int():
    """
    Test ValueError is raised when length of y is less than window_size
    when offset is an int.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 6,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                 )
    y = pd.Series(np.arange(10))

    err_msg = re.escape(
        "Length of `y` must be greater than the maximum window size "
        "needed by the forecaster. This is because  "
        "the offset (6) is larger than the available "
        "data. Try to decrease the size of the offset (6), "
        "the number of `n_offsets` (2) or increase the "
        "size of `y`.\n"
        "    Length `y`: 10.\n"
        "    Max window size: 12.\n"
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.fit(y=y)


@pytest.mark.parametrize("offset, y", 
                         [({'days': 6}, 
                           pd.Series(np.arange(10),
                                     index=pd.date_range(start='01/01/2021', periods=10, freq='D'))), 
                          ({'months': 6}, 
                           pd.Series(np.arange(10), 
                                     index=pd.date_range(start='01/01/2021', periods=10, freq='MS')))])
def test_fit_ValueError_length_y_less_than_window_size_offset_DateOffset(offset, y):
    """
    Test ValueError is raised when length of y is less than window_size
    when offset is a pandas DateOffset.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = DateOffset(**offset),
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )

    err_msg = re.escape(
        f"The length of `y` (10), must be greater than or equal "
        f"to the window size ({forecaster.window_size}). This is because  "
        f"the offset ({forecaster.offset}) is larger than the available "
        f"data. Try to decrease the size of the offset ({forecaster.offset}), "
        f"the number of `n_offsets` (2) or increase the "
        f"size of `y`."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.fit(y=y)


def test_fit_y_index_DatetimeIndex():
    """
    Test index_freq_ is set correctly when y index is a DatetimeIndex.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )
    y = pd.Series(
        np.random.rand(10), index=pd.date_range(start='1/1/2021', periods=10)
    )
    forecaster.fit(y)

    assert forecaster.index_freq_ == y.index.freqstr


def test_fit_y_index_not_DatetimeIndex():
    """
    Test index_freq_ is set correctly when y index is not a DatetimeIndex.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )
    y = pd.Series(np.random.rand(10))
    forecaster.fit(y)

    assert forecaster.index_freq_ == y.index.step


def test_fit_offset_int():
    """
    Test window_size and last_window_ are set correctly when offset is an int.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )
    y = pd.Series(np.random.rand(10), index=pd.RangeIndex(start=0, stop=10))
    forecaster.fit(y)

    assert forecaster.window_size == 4.0
    assert forecaster.last_window_.equals(y)


def test_fit_offset_DateOffset():
    """
    Test window_size and last_window_ are set correctly when offset is a DateOffset.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = DateOffset(days=2),
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )
    y = pd.Series(np.random.rand(10), index=pd.date_range(start='01/01/2021', periods=10))
    forecaster.fit(y)

    assert forecaster.window_size == 4.0
    assert forecaster.last_window_.equals(y)


def test_fit_in_sample_residuals_stored():
    """
    Test that values of in_sample_residuals_ are stored after fitting.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 1,
                     n_offsets     = 1,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )
    forecaster.fit(y=pd.Series(np.arange(5)), store_in_sample_residuals=True)
    results = forecaster.in_sample_residuals_
    expected = np.array([1., 1., 1., 1.])

    assert isinstance(results, np.ndarray)
    np.testing.assert_array_almost_equal(results, expected)


def test_fit_same_residuals_when_residuals_greater_than_10000():
    """
    Test fit return same residuals when residuals len is greater than 10_000.
    Testing with two different forecaster.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 3,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )
    forecaster.fit(y=pd.Series(np.arange(12_000)), store_in_sample_residuals=True)
    results_1 = forecaster.in_sample_residuals_
    forecaster = ForecasterEquivalentDate(
                     offset        = 3,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     forecaster_id = None
                 )
    forecaster.fit(y=pd.Series(np.arange(12_000)), store_in_sample_residuals=True)
    results_2 = forecaster.in_sample_residuals_
    
    assert isinstance(results_1, np.ndarray)
    assert isinstance(results_2, np.ndarray)
    assert len(results_1) == 10_000
    assert len(results_2) == 10_000
    np.testing.assert_array_almost_equal(results_1, results_2)


def test_fit_in_sample_residuals_by_bin_stored_offset_int():
    """
    Test that values of in_sample_residuals_by_bin are stored after fitting
    when offset is an int.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = 2,
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     binner_kwargs = {'n_bins': 3},
                     forecaster_id = None
                 )
    forecaster.fit(y, store_in_sample_residuals=True)

    expected_1 = np.array([
        0.25780865,  0.00437941,  0.50760399,  0.19761913, -0.36918468,
       -0.16185058, -0.38767003,  0.19057608,  0.02651728, -0.50090572,
        0.00716913,  0.34363161, -0.23581652, -0.22338489,  0.24128337,
        0.07510401,  0.27737941,  0.49579211,  0.14147915, -0.07960618,
        0.04301524, -0.40726874, -0.36166069, -0.23872798, -0.24840197,
        0.35536505, -0.23564641,  0.00408149,  0.23795327, -0.03865355,
        0.16434644, -0.15143191, -0.00199522,  0.490416  ,  0.51806922,
       -0.10098851, -0.06130272, -0.58199452, -0.46677101,  0.10609867,
        0.39568995, -0.01476693, -0.10876306,  0.652919  , -0.15518659,
       -0.00511305
    ])

    expected_2 = {
        0: np.array([ 0.02651728,  0.00716913,  0.34363161, -0.22338489,  0.24128337,
            0.27737941,  0.49579211,  0.35536505, -0.23564641,  0.23795327,
            0.16434644,  0.490416  ,  0.10609867, -0.01476693,  0.652919  ]),
        1: np.array([ 0.25780865,  0.00437941,  0.50760399,  0.19761913,  0.19057608,
            -0.23581652,  0.07510401, -0.23872798, -0.24840197,  0.00408149,
            -0.03865355, -0.15143191, -0.00199522,  0.51806922,  0.39568995]),
        2: np.array([-0.36918468, -0.16185058, -0.38767003, -0.50090572,  0.14147915,
            -0.07960618,  0.04301524, -0.40726874, -0.36166069, -0.10098851,
            -0.06130272, -0.58199452, -0.46677101, -0.10876306, -0.15518659,
            -0.00511305])
    }

    expected_3 = {
        0: (0.192909495, 0.41830825),
        1: (0.41830825, 0.5539681),
        2: (0.5539681, 0.850116585),
    }

    np.testing.assert_array_almost_equal(
        forecaster.in_sample_residuals_, expected_1
    )
    for k in expected_2.keys():
        np.testing.assert_array_almost_equal(forecaster.in_sample_residuals_by_bin_[k], expected_2[k])
    for k in expected_3.keys():
        assert forecaster.binner_intervals_[k][0] == approx(expected_3[k][0])
        assert forecaster.binner_intervals_[k][1] == approx(expected_3[k][1])


def test_fit_in_sample_residuals_by_bin_stored_offset_DateOffset():
    """
    Test that values of in_sample_residuals_by_bin are stored after fitting
    when offset is a DateOffset.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = DateOffset(days=2),
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     binner_kwargs = {'n_bins': 3},
                     forecaster_id = None
                 )
    forecaster.fit(y, store_in_sample_residuals=True)

    expected_1 = np.array([
        0.25780865,  0.00437941,  0.50760399,  0.19761913, -0.36918468,
       -0.16185058, -0.38767003,  0.19057608,  0.02651728, -0.50090572,
        0.00716913,  0.34363161, -0.23581652, -0.22338489,  0.24128337,
        0.07510401,  0.27737941,  0.49579211,  0.14147915, -0.07960618,
        0.04301524, -0.40726874, -0.36166069, -0.23872798, -0.24840197,
        0.35536505, -0.23564641,  0.00408149,  0.23795327, -0.03865355,
        0.16434644, -0.15143191, -0.00199522,  0.490416  ,  0.51806922,
       -0.10098851, -0.06130272, -0.58199452, -0.46677101,  0.10609867,
        0.39568995, -0.01476693, -0.10876306,  0.652919  , -0.15518659,
       -0.00511305
    ])

    expected_2 = {
        0: np.array([ 0.02651728,  0.00716913,  0.34363161, -0.22338489,  0.24128337,
            0.27737941,  0.49579211,  0.35536505, -0.23564641,  0.23795327,
            0.16434644,  0.490416  ,  0.10609867, -0.01476693,  0.652919  ]),
        1: np.array([ 0.25780865,  0.00437941,  0.50760399,  0.19761913,  0.19057608,
            -0.23581652,  0.07510401, -0.23872798, -0.24840197,  0.00408149,
            -0.03865355, -0.15143191, -0.00199522,  0.51806922,  0.39568995]),
        2: np.array([-0.36918468, -0.16185058, -0.38767003, -0.50090572,  0.14147915,
            -0.07960618,  0.04301524, -0.40726874, -0.36166069, -0.10098851,
            -0.06130272, -0.58199452, -0.46677101, -0.10876306, -0.15518659,
            -0.00511305])
    }

    expected_3 = {
        0: (0.192909495, 0.41830825),
        1: (0.41830825, 0.5539681),
        2: (0.5539681, 0.850116585),
    }

    np.testing.assert_array_almost_equal(
        forecaster.in_sample_residuals_, expected_1
    )
    for k in expected_2.keys():
        np.testing.assert_array_almost_equal(forecaster.in_sample_residuals_by_bin_[k], expected_2[k])
    for k in expected_3.keys():
        assert forecaster.binner_intervals_[k][0] == approx(expected_3[k][0])
        assert forecaster.binner_intervals_[k][1] == approx(expected_3[k][1])


def test_fit_in_sample_residuals_not_stored_probabilistic_mode_binned():
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False`. Binner intervals are stored.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = DateOffset(days=2),
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     binner_kwargs = {'n_bins': 3},
                     forecaster_id = None
                 )
    forecaster.fit(y, store_in_sample_residuals=False)

    expected_binner_intervals_ = {
        0: (0.192909495, 0.41830825),
        1: (0.41830825, 0.5539681),
        2: (0.5539681, 0.850116585),
    }

    assert forecaster.in_sample_residuals_ is None
    assert forecaster.in_sample_residuals_by_bin_ is None
    assert forecaster.binner_intervals_.keys() == expected_binner_intervals_.keys()
    for k in expected_binner_intervals_.keys():
        assert forecaster.binner_intervals_[k][0] == approx(expected_binner_intervals_[k][0])
        assert forecaster.binner_intervals_[k][1] == approx(expected_binner_intervals_[k][1])


def test_fit_in_sample_residuals_not_stored_probabilistic_mode_False():
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False` and _probabilistic_mode=False.
    """
    forecaster = ForecasterEquivalentDate(
                     offset        = DateOffset(days=2),
                     n_offsets     = 2,
                     agg_func      = np.mean,
                     binner_kwargs = {'n_bins': 3},
                     forecaster_id = None
                 )
    forecaster._probabilistic_mode = False
    forecaster.fit(y=y, store_in_sample_residuals=False)

    assert forecaster.in_sample_residuals_ is None
    assert forecaster.in_sample_residuals_by_bin_ is None
    assert forecaster.binner_intervals_ is None



# Unit test select_n_jobs_backtesting
# ==============================================================================
import pytest
from joblib import cpu_count
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor
from skforecast.model_selection._utils import select_n_jobs_backtesting
from skforecast.recursive import ForecasterRecursive
from skforecast.direct import ForecasterDirect
from skforecast.recursive import ForecasterRecursiveMultiSeries
from skforecast.sarimax import Sarimax
from skforecast.recursive import ForecasterSarimax
from skforecast.recursive import ForecasterEquivalentDate


@pytest.mark.parametrize("forecaster, refit, n_jobs_expected", 
    [(ForecasterRecursive(LinearRegression(), lags=2), True, 1),
     (ForecasterRecursive(make_pipeline(StandardScaler(), LinearRegression()), lags=2), True, 1),
     (ForecasterRecursive(LinearRegression(), lags=2), 2, 1),
     (ForecasterRecursive(LinearRegression(), lags=2), False, 1),
     (ForecasterRecursive(HistGradientBoostingRegressor(), lags=2), 1, cpu_count() - 1),
     (ForecasterRecursive(HistGradientBoostingRegressor(), lags=2), 2, 1),
     (ForecasterRecursive(HistGradientBoostingRegressor(), lags=2), 0, cpu_count() - 1),
     (ForecasterRecursive(LGBMRegressor(n_jobs=1), lags=2), True, cpu_count() - 1),
     (ForecasterRecursive(LGBMRegressor(n_jobs=1), lags=2), False, cpu_count() - 1),
     (ForecasterRecursive(LGBMRegressor(), lags=2), True, 1),
     (ForecasterRecursive(LGBMRegressor(), lags=2), False, 1),
     (ForecasterDirect(LinearRegression(), steps=3, lags=2), True, 1),
     (ForecasterDirect(LinearRegression(), steps=3, lags=2), 2, 1),
     (ForecasterDirect(LinearRegression(), steps=3, lags=2), False, 1),
     (ForecasterDirect(HistGradientBoostingRegressor(), steps=3, lags=2), 1, 1),
     (ForecasterDirect(HistGradientBoostingRegressor(), steps=3, lags=2), 2, 1),
     (ForecasterDirect(HistGradientBoostingRegressor(), steps=3, lags=2), 0, 1),
     (ForecasterDirect(LGBMRegressor(), steps=3, lags=2), True, 1),
     (ForecasterDirect(LGBMRegressor(), steps=3, lags=2), False, 1),
     (ForecasterRecursiveMultiSeries(LinearRegression(), lags=2), True, cpu_count() - 1),
     (ForecasterRecursiveMultiSeries(LinearRegression(), lags=2), 2, 1),
     (ForecasterRecursiveMultiSeries(LinearRegression(), lags=2), False, cpu_count() - 1),
     (ForecasterRecursiveMultiSeries(HistGradientBoostingRegressor(), lags=2), 1, cpu_count() - 1),
     (ForecasterRecursiveMultiSeries(HistGradientBoostingRegressor(), lags=2), 2, 1),
     (ForecasterRecursiveMultiSeries(HistGradientBoostingRegressor(), lags=2), 0, cpu_count() - 1),
     (ForecasterRecursiveMultiSeries(LGBMRegressor(n_jobs=1), lags=2), True, cpu_count() - 1),
     (ForecasterRecursiveMultiSeries(LGBMRegressor(n_jobs=1), lags=2), False, cpu_count() - 1),
     (ForecasterRecursiveMultiSeries(LGBMRegressor(), lags=2), True, 1),
     (ForecasterRecursiveMultiSeries(LGBMRegressor(), lags=2), False, 1),
     (ForecasterSarimax(Sarimax((1, 0, 1))), True, 1),
     (ForecasterSarimax(Sarimax((1, 0, 1))), 2, 1),
     (ForecasterSarimax(Sarimax((1, 0, 1))), False, 1),
     (ForecasterEquivalentDate(2), True, 1),
     (ForecasterEquivalentDate(2), 2, 1),
     (ForecasterEquivalentDate(2), False, 1)
], 
ids=lambda info: f'info: {info}')
def test_select_n_jobs_backtesting(forecaster, refit, n_jobs_expected):
    """
    Test select_n_jobs_backtesting
    """
    n_jobs = select_n_jobs_backtesting(
                 forecaster = forecaster, 
                 refit      = refit
             )
    
    assert n_jobs == n_jobs_expected



# Unit test backtesting ForecasterEquivalentDate
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from skforecast.recursive import ForecasterEquivalentDate
from skforecast.model_selection._validation import backtesting_forecaster
from skforecast.model_selection._split import TimeSeriesFold

# Fixtures
from .fixtures_forecaster_equivalent_date import y


@pytest.mark.parametrize("initial_train_size", 
                         [len(y) - 20, "2000-01-30 00:00:00"],
                         ids=lambda init: f'initial_train_size: {init}')
def test_backtesting_with_ForecasterEquivalentDate(initial_train_size):
    """
    Test backtesting with ForecasterEquivalentDate.
    """
    forecaster = ForecasterEquivalentDate(
                     offset    = pd.DateOffset(days=10),
                     n_offsets = 2 
                 )
    cv = TimeSeriesFold(
             initial_train_size = initial_train_size,
             steps              = 5,
             refit              = True,
         )

    metric, predictions = backtesting_forecaster(
        forecaster = forecaster,
        y          = y,
        cv         = cv,
        metric     = 'mean_absolute_error',
        verbose    = False,
        n_jobs     = 'auto'
    )

    expected_metric = pd.DataFrame({'mean_absolute_error': [0.2537094475]})
    expected = pd.DataFrame(
        data    = np.array([0.48878949, 0.78924075, 0.58151378, 0.3353507, 0.56024382,
                            0.53047716, 0.27214019, 0.20185749, 0.41263271, 0.58140185,
                            0.36325295, 0.64156648, 0.57765904, 0.5523543, 0.57413684,
                            0.31761006, 0.39406999, 0.56082619, 0.61893703, 0.5664064]),
        index   = pd.date_range(start='2000-01-31', periods=20, freq='D'),
        columns = ['pred']
    )

    pd.testing.assert_frame_equal(metric, expected_metric)
    pd.testing.assert_frame_equal(predictions, expected)


@pytest.mark.parametrize("initial_train_size", 
                         [len(y) - 20, "2000-01-30 23:59:00"],
                         ids=lambda init: f'initial_train_size: {init}')
def test_backtesting_with_ForecasterEquivalentDate_interval(initial_train_size):
    """
    Test backtesting with ForecasterEquivalentDate and interval prediction.
    """
    forecaster = ForecasterEquivalentDate(
                     offset    = pd.DateOffset(days=3),
                     n_offsets = 3 
                 )
    cv = TimeSeriesFold(
             initial_train_size = initial_train_size,
             steps              = 3,
             refit              = False,
         )

    metric, predictions = backtesting_forecaster(
        forecaster      = forecaster,
        y               = y,
        cv              = cv,
        metric          = 'mean_absolute_error',
        interval        = [10, 90],
        interval_method = 'conformal',
        verbose         = False,
        n_jobs          = 'auto'
    )
    
    expected_metric = pd.DataFrame({'mean_absolute_error': [0.2273233665]})
    expected = pd.DataFrame(
                   data = np.array([
                              [0.60004613,  0.1257829 ,  1.07430936],
                              [0.44704276,  0.240286  ,  0.65379952],
                              [0.5345961 ,  0.34670492,  0.72248727],
                              [0.34760385, -0.09290349,  0.78811119],
                              [0.35012471, -0.09038263,  0.79063205],
                              [0.47454251,  0.32686905,  0.62221597],
                              [0.27135109, -0.16915625,  0.71185843],
                              [0.38441517,  0.06673199,  0.70209835],
                              [0.45803337,  0.25127661,  0.66479012],
                              [0.33738045, -0.10312689,  0.77788779],
                              [0.58430687,  0.11004364,  1.0585701 ],
                              [0.562428  ,  0.35692325,  0.76793275],
                              [0.4739577 ,  0.32628424,  0.62163116],
                              [0.64772413,  0.45862738,  0.83682089],
                              [0.45734655,  0.25058979,  0.6641033 ],
                              [0.41515782,  0.20840107,  0.62191458],
                              [0.64405611,  0.45495935,  0.83315286],
                              [0.64202919,  0.45293244,  0.83112595],
                              [0.35652584, -0.0839815 ,  0.79703318],
                              [0.50727114,  0.35959768,  0.6549446 ]]),
                   index = pd.date_range(start='2000-01-31', periods=20, freq='D'),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )

    pd.testing.assert_frame_equal(metric, expected_metric)
    pd.testing.assert_frame_equal(predictions, expected)



# Unit test _evaluate_grid_hyperparameters
# ==============================================================================
import re
import os
import pytest
import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error, mean_absolute_percentage_error
from sklearn.model_selection import ParameterGrid
from sklearn.linear_model import Ridge
from sklearn.preprocessing import StandardScaler
from skforecast.metrics import mean_absolute_scaled_error, root_mean_squared_scaled_error
from skforecast.exceptions import OneStepAheadValidationWarning
from skforecast.recursive import ForecasterRecursive
from skforecast.direct import ForecasterDirect
from skforecast.model_selection._search import _evaluate_grid_hyperparameters
from skforecast.model_selection._split import TimeSeriesFold, OneStepAheadFold
from skforecast.preprocessing import RollingFeatures

# Fixtures
from ..fixtures_model_selection import y
from ..fixtures_model_selection import y_feature_selection
from ..fixtures_model_selection import exog_feature_selection

from tqdm import tqdm
from functools import partialmethod
tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)  # hide progress bar


def test_TypeError_evaluate_grid_hyperparameters_when_cv_not_valid():
    """
    Test TypeError is raised in _evaluate_grid_hyperparameters when cv is not
    a valid splitter.
    """
    class DummyCV:
        pass

    cv = DummyCV()
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    
    err_msg = re.escape(
        f"`cv` must be an instance of `TimeSeriesFold` or `OneStepAheadFold`. "
        f"Got {type(cv)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        _evaluate_grid_hyperparameters(
            forecaster  = forecaster,
            y           = y,
            cv          = cv,
            lags_grid   = [2, 4],
            param_grid  = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
            metric      = 'mean_absolute_error',
            return_best = True,
            verbose     = False
        )


def test_ValueError_evaluate_grid_hyperparameters_when_return_best_and_len_y_exog_different():
    """
    Test ValueError is raised in _evaluate_grid_hyperparameters when return_best 
    and length of `y` and `exog` do not match.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    exog = y[:30]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y[:-12]),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    err_msg = re.escape(
        (f"`exog` must have same number of samples as `y`. "
         f"length `exog`: ({len(exog)}), length `y`: ({len(y)})")
    )
    with pytest.raises(ValueError, match = err_msg):
        _evaluate_grid_hyperparameters(
            forecaster  = forecaster,
            y           = y,
            exog        = exog,
            cv          = cv,
            lags_grid   = [2, 4],
            param_grid  = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
            metric      = 'mean_absolute_error',
            return_best = True,
            verbose     = False
        )


def test_ValueError_evaluate_grid_hyperparameters_metric_list_duplicate_names():
    """
    Test ValueError is raised in _evaluate_grid_hyperparameters when a `list` of 
    metrics is used with duplicate names.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y[:-12]),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    err_msg = re.escape('When `metric` is a `list`, each metric name must be unique.')
    with pytest.raises(ValueError, match = err_msg):
        _evaluate_grid_hyperparameters(
            forecaster  = forecaster,
            y           = y,
            cv          = cv,
            lags_grid   = [2, 4],
            param_grid  = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}],
            metric      = ['mean_absolute_error', mean_absolute_error],
            return_best = False,
            verbose     = False
        )


def test_output_evaluate_grid_hyperparameters_ForecasterRecursive_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterRecursive with mocked
    (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid) * len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster  = forecaster,
                  y           = y,
                  cv          = cv,
                  lags_grid   = lags_grid,
                  param_grid  = param_grid,
                  metric      = 'mean_squared_error',
                  return_best = False,
                  verbose     = False
              )
    
    expected_results = pd.DataFrame(
        {
            "lags": [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
            "lags_label": [
                [1, 2],
                [1, 2],
                [1, 2],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
            ],
            "params": [
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
            ],
            "mean_squared_error": np.array(
                [0.06464646, 0.06502362, 0.06745534, 0.06779272, 0.06802481, 0.06948609]
            ),
            "alpha": np.array([0.01, 0.1, 1.0, 0.01, 0.1, 1.0]),
        },
        index=pd.RangeIndex(start=0, stop=idx, step=1),
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterRecursive_with_window_features():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterRecursive with 
    window features (mocked done in Skforecast v0.14.0).
    """
    window_features = RollingFeatures(
        stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation'],
        window_sizes = 3,
    )
    forecaster = ForecasterRecursive(
                     regressor       = Ridge(random_state=123),
                     lags            = 2, 
                     window_features = window_features,
                 )
    
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    results = _evaluate_grid_hyperparameters(
                  forecaster  = forecaster,
                  y           = y,
                  cv          = cv,
                  lags_grid   = lags_grid,
                  param_grid  = param_grid,
                  metric      = 'mean_squared_error',
                  return_best = False,
                  verbose     = False
              )
    
    expected_results = pd.DataFrame(
        {'lags': [np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2])],
        'lags_label': [np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2])],
        'params': [{'alpha': 1},
                {'alpha': 1},
                {'alpha': 0.1},
                {'alpha': 0.1},
                {'alpha': 0.01},
                {'alpha': 0.01}],
        'mean_squared_error': [0.08284287269722487,
                0.0832323925238072,
                0.0908332414354649,
                0.09264625639393599,
                0.10408604464751903,
                0.10624605486182163],
        'alpha': [1.0, 1.0, 0.1, 0.1, 0.01, 0.01]}
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterRecursive_with_differentiation_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterRecursive with mocked
    (mocked done in Skforecast v0.10.0) when differentiation is used.
    """
    forecaster = ForecasterRecursive(
                     regressor       = Ridge(random_state=123),
                     lags            = 2,
                     differentiation = 1
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = 1,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )

    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid) * len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster  = forecaster,
                  y           = y,
                  cv          = cv,
                  lags_grid   = lags_grid,
                  param_grid  = param_grid,
                  metric      = 'mean_squared_error',
                  return_best = False,
                  verbose     = False
              ).reset_index(drop=True)

    expected_results = pd.DataFrame(
        {
            "lags": [[1, 2, 3, 4], [1, 2], [1, 2, 3, 4], [1, 2], [1, 2], [1, 2, 3, 4]],
            "lags_label": [
                [1, 2, 3, 4],
                [1, 2],
                [1, 2, 3, 4],
                [1, 2],
                [1, 2],
                [1, 2, 3, 4],
            ],
            "params": [
                {"alpha": 1},
                {"alpha": 1},
                {"alpha": 0.1},
                {"alpha": 0.1},
                {"alpha": 0.01},
                {"alpha": 0.01},
            ],
            "mean_squared_error": np.array(
                [0.09168123, 0.09300068, 0.09930084, 0.09960109, 0.10102995, 0.1012931]
            ),
            "alpha": np.array([1.0, 1.0, 0.1, 0.1, 0.01, 0.01]),
        },
        index=pd.RangeIndex(start=0, stop=idx, step=1),
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterRecursive_lags_grid_dict_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterRecursive when 
    `lags_grid` is a dict with mocked (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )
    lags_grid = {'lags_1': 2, 'lags_2': 4}
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid) * len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster  = forecaster,
                  y           = y,
                  cv          = cv,
                  lags_grid   = lags_grid,
                  param_grid  = param_grid,
                  metric      = 'mean_squared_error',
                  return_best = False,
                  verbose     = False
              )

    expected_results = pd.DataFrame(
        {
            "lags": [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
            "lags_label": ["lags_1", "lags_1", "lags_1", "lags_2", "lags_2", "lags_2"],
            "params": [
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
            ],
            "mean_squared_error": np.array(
                [0.06464646, 0.06502362, 0.06745534, 0.06779272, 0.06802481, 0.06948609]
            ),
            "alpha": np.array([0.01, 0.1, 1.0, 0.01, 0.1, 1.0]),
        },
        index=pd.RangeIndex(start=0, stop=idx, step=1),
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterRecursive_lags_grid_is_None_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterRecursive when 
    `lags_grid` is None with mocked (mocked done in Skforecast v0.4.3), 
    should use forecaster.lags as lags_grid.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )
    lags_grid = None
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    results = _evaluate_grid_hyperparameters(
                  forecaster  = forecaster,
                  y           = y,
                  cv          = cv,
                  lags_grid   = lags_grid,
                  param_grid  = param_grid,
                  metric      = 'mean_squared_error',
                  return_best = False,
                  verbose     = False
              )

    expected_results = pd.DataFrame(
        {
            "lags": [[1, 2], [1, 2], [1, 2]],
            "lags_label": [[1, 2], [1, 2], [1, 2]],
            "params": [{"alpha": 0.01}, {"alpha": 0.1}, {"alpha": 1}],
            "mean_squared_error": np.array([0.06464646, 0.06502362, 0.06745534]),
            "alpha": np.array([0.01, 0.1, 1.0]),
        },
        index=pd.RangeIndex(start=0, stop=3, step=1),
    )

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_evaluate_grid_hyperparameters_ForecasterRecursive_metric_list_with_mocked():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterRecursive with mocked
    and multiple metrics (mocked done in Skforecast v0.4.3).
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    idx = len(lags_grid) * len(param_grid)

    results = _evaluate_grid_hyperparameters(
                  forecaster  = forecaster,
                  y           = y,
                  cv          = cv,
                  lags_grid   = lags_grid,
                  param_grid  = param_grid,
                  metric      = ['mean_squared_error', mean_absolute_error],
                  return_best = False,
                  verbose     = False
              )

    expected_results = pd.DataFrame(
        {
            "lags": [[1, 2], [1, 2], [1, 2], [1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]],
            "lags_label": [
                [1, 2],
                [1, 2],
                [1, 2],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
                [1, 2, 3, 4],
            ],
            "params": [
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
                {"alpha": 0.01},
                {"alpha": 0.1},
                {"alpha": 1},
            ],
            "mean_squared_error": np.array(
                [0.06464646, 0.06502362, 0.06745534, 0.06779272, 0.06802481, 0.06948609]
            ),
            "mean_absolute_error": np.array(
                [0.20278812, 0.20314819, 0.20519952, 0.20601567, 0.206323, 0.20747017]
            ),
            "alpha": np.array([0.01, 0.1, 1.0, 0.01, 0.1, 1.0]),
        },
        index=pd.RangeIndex(start=0, stop=idx, step=1),
    )

    pd.testing.assert_frame_equal(results, expected_results)


@pytest.mark.parametrize("lags_grid", 
                         [[2, 4], {'lags_1': 2, 'lags_2': 4}], 
                         ids=lambda lg: f'lags_grid: {lg}')
def test_evaluate_grid_hyperparameters_when_return_best_ForecasterRecursive(lags_grid):
    """
    Test forecaster is refitted when return_best=True in 
    _evaluate_grid_hyperparameters with ForecasterRecursive.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    _evaluate_grid_hyperparameters(
        forecaster  = forecaster,
        y           = y,
        cv          = cv,
        lags_grid   = lags_grid,
        param_grid  = param_grid,
        metric      = 'mean_squared_error',
        return_best = True,
        verbose     = False
    )
    
    expected_lags = np.array([1, 2])
    expected_alpha = 0.01
    
    assert (expected_lags == forecaster.lags).all()
    assert expected_alpha == forecaster.regressor.alpha


@pytest.mark.parametrize("lags_grid", 
                         [[2, 4], {'lags_1': 2, 'lags_2': 4}], 
                         ids=lambda lg: f'lags_grid: {lg}')
def test_evaluate_grid_hyperparameters_when_return_best_and_list_metrics(lags_grid):
    """
    Test forecaster is refitted when return_best=True in _evaluate_grid_hyperparameters
    and multiple metrics.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    _evaluate_grid_hyperparameters(
        forecaster    = forecaster,
        y             = y,
        cv            = cv,
        lags_grid     = lags_grid,
        param_grid    = param_grid,
        metric        = [mean_absolute_percentage_error, 'mean_squared_error'],
        return_best   = True,
        verbose       = False,
        show_progress = False
    )
    
    expected_lags = np.array([1, 2])
    expected_alpha = 1.
    
    assert (expected_lags == forecaster.lags).all()
    assert expected_alpha == forecaster.regressor.alpha


def test_evaluate_grid_hyperparameters_output_file_when_single_metric():
    """
    Test output file is created when output_file is passed to
    _evaluate_grid_hyperparameters and single metric.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False
         )
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    output_file = 'test_evaluate_grid_hyperparameters_output_file.txt'

    results = _evaluate_grid_hyperparameters(
                  forecaster    = forecaster,
                  y             = y,
                  cv            = cv,
                  lags_grid     = lags_grid,
                  param_grid    = param_grid,
                  metric        = 'mean_squared_error',
                  return_best   = False,
                  verbose       = False,
                  show_progress = False,
                  output_file   = output_file
              )
    results  = results.astype({'lags': str, 'lags_label': str, 'params': str})

    assert os.path.isfile(output_file)
    output_file_content = pd.read_csv(output_file, sep='\t', low_memory=False)
    output_file_content = output_file_content.sort_values(by='mean_squared_error')
    output_file_content = output_file_content.astype({'lags': str, 'lags_label': str, 'params': str})
    pd.testing.assert_frame_equal(results, output_file_content)
    os.remove(output_file)


def test_evaluate_grid_hyperparameters_output_file_when_single_metric_as_list():
    """ 
    Test output file is created when output_file is passed to 
    _evaluate_grid_hyperparameters and single metric as list.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )

    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    output_file = 'test_evaluate_grid_hyperparameters_output_file.txt'
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  cv                 = cv,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  metric             = ['mean_squared_error'],
                  return_best        = False,
                  verbose            = False,
                  show_progress      = False,
                  output_file        = output_file
              )
    results  = results.astype({'lags': str, 'lags_label': str, 'params': str})

    assert os.path.isfile(output_file)
    output_file_content = pd.read_csv(output_file, sep='\t', low_memory=False)
    output_file_content = output_file_content.sort_values(by='mean_squared_error')
    output_file_content = output_file_content.astype({'lags': str, 'lags_label': str, 'params': str})
    pd.testing.assert_frame_equal(results, output_file_content)
    os.remove(output_file)


def test_evaluate_grid_hyperparameters_output_file_when_2_metrics_as_list():
    """
    Test output file is created when output_file is passed to 
    _evaluate_grid_hyperparameters and 2 metrics as list.
    """
    forecaster = ForecasterRecursive(
                     regressor = Ridge(random_state=123),
                     lags      = 2 
                 )

    n_validation = 12
    y_train = y[:-n_validation]
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]
    output_file = 'test_evaluate_grid_hyperparameters_output_file.txt'

    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    results = _evaluate_grid_hyperparameters(
                  forecaster         = forecaster,
                  y                  = y,
                  cv                 = cv,
                  lags_grid          = lags_grid,
                  param_grid         = param_grid,
                  metric             = ['mean_squared_error', 'mean_absolute_error'],
                  return_best        = False,
                  verbose            = False,
                  show_progress      = False,
                  output_file        = output_file
              )
    results  = results.astype({'lags': str, 'lags_label': str, 'params': str})

    assert os.path.isfile(output_file)
    output_file_content = pd.read_csv(output_file, sep='\t', low_memory=False)
    output_file_content = output_file_content.sort_values(by='mean_squared_error')
    output_file_content = output_file_content.astype({'lags': str, 'lags_label': str, 'params': str})
    pd.testing.assert_frame_equal(results, output_file_content)
    os.remove(output_file)


def test_output_evaluate_grid_hyperparameters_ForecasterDirect_with_window_features():
    """
    Test output of _evaluate_grid_hyperparameters in ForecasterDirect with 
    window features (mocked done in Skforecast v0.14.0).
    """
    window_features = RollingFeatures(
        stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation'],
        window_sizes = 3,
    )
    forecaster = ForecasterDirect(
                     regressor       = Ridge(random_state=123),
                     steps           = 3,
                     lags            = 2, 
                     window_features = window_features,
                 )
    
    n_validation = 12
    y_train = y[:-n_validation]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    lags_grid = [2, 4]
    param_grid = [{'alpha': 0.01}, {'alpha': 0.1}, {'alpha': 1}]

    results = _evaluate_grid_hyperparameters(
                  forecaster  = forecaster,
                  y           = y,
                  cv          = cv,
                  lags_grid   = lags_grid,
                  param_grid  = param_grid,
                  metric      = 'mean_squared_error',
                  return_best = False,
                  verbose     = False
              )
    
    expected_results = pd.DataFrame(
        {'lags': [np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2])],
        'lags_label': [np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2]),
                np.array([1, 2, 3, 4]),
                np.array([1, 2])],
        'params': [{'alpha': 1},
                {'alpha': 1},
                {'alpha': 0.1},
                {'alpha': 0.1},
                {'alpha': 0.01},
                {'alpha': 0.01}],
        'mean_squared_error': [0.07424336814199613,
                0.07576112822191301,
                0.07677225532446876,
                0.08046370393440734,
                0.08411584414110274,
                0.08911485459058603],
        'alpha': [1.0, 1.0, 0.1, 0.1, 0.01, 0.01]}
    )

    pd.testing.assert_frame_equal(results, expected_results)


@pytest.mark.parametrize("initial_train_size", 
                         [100, '2020-01-05 03:00:00', pd.to_datetime('2020-01-05 03:00:00')], 
                         ids=lambda initial_train_size: f'initial_train_size: {initial_train_size}')
@pytest.mark.parametrize(
        "forecaster",
        [
            ForecasterRecursive(
                regressor=Ridge(random_state=678),
                lags=3,
                transformer_y=None,
                forecaster_id='Recursive_no_transformer'
            ),
            ForecasterDirect(
                regressor=Ridge(random_state=678),
                steps=1,
                lags=3,
                transformer_y=None,
                forecaster_id='Direct_no_transformer'
            ),
            ForecasterRecursive(
                regressor=Ridge(random_state=678),
                lags=3,
                transformer_y=StandardScaler(),
                transformer_exog=StandardScaler(),
                forecaster_id='Recursive_transformers'
            ),
            ForecasterDirect(
                regressor=Ridge(random_state=678),
                steps=1,
                lags=3,
                transformer_y=StandardScaler(),
                transformer_exog=StandardScaler(),
                forecaster_id='Direct_transformer'
            )
        ],
ids=lambda forecaster: f'forecaster: {forecaster.forecaster_id}')
def test_evaluate_grid_hyperparameters_equivalent_outputs_backtesting_one_step_ahead(
    forecaster, initial_train_size
):
    """
    Test that the outputs of _evaluate_grid_hyperparameters are equivalent when
    using backtesting and one-step-ahead.
    """
    metrics = [
        "mean_absolute_error",
        "mean_squared_error",
        mean_absolute_percentage_error,
        mean_absolute_scaled_error,
        root_mean_squared_scaled_error,
    ]
    param_grid = {
        "alpha": np.logspace(-3, 3, 2),
    }
    lags_grid = [3, 5, 7]
    param_grid = list(ParameterGrid(param_grid))
    cv_backtesnting = TimeSeriesFold(
            steps                 = 1,
            initial_train_size    = initial_train_size,
            window_size           = None,
            differentiation       = None,
            refit                 = False,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    cv_one_step_ahead = OneStepAheadFold(
            initial_train_size    = initial_train_size,
            return_all_indexes    = False,
        )
    
    results_backtesting = _evaluate_grid_hyperparameters(
        forecaster         = forecaster,
        y                  = y_feature_selection,
        exog               = exog_feature_selection,
        cv                 = cv_backtesnting,
        param_grid         = param_grid,
        lags_grid          = lags_grid,
        metric             = metrics,
        return_best        = False,
        n_jobs             = 'auto',
        verbose            = False,
        show_progress      = False
    )

    warn_msg = re.escape(
        "One-step-ahead predictions are used for faster model comparison, but they "
        "may not fully represent multi-step prediction performance. It is recommended "
        "to backtest the final model for a more accurate multi-step performance "
        "estimate."
    )
    with pytest.warns(OneStepAheadValidationWarning, match = warn_msg):
        results_one_step_ahead = _evaluate_grid_hyperparameters(
            forecaster         = forecaster,
            y                  = y_feature_selection,
            exog               = exog_feature_selection,
            cv                 = cv_one_step_ahead,
            param_grid         = param_grid,
            lags_grid          = lags_grid,
            metric             = metrics,
            return_best        = False,
            verbose            = False,
            show_progress      = False
        )

    pd.testing.assert_frame_equal(results_backtesting, results_one_step_ahead)



# Unit test select_n_jobs_fit_forecaster
# ==============================================================================
import pytest
from joblib import cpu_count
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor
from skforecast.utils.utils import select_n_jobs_fit_forecaster


@pytest.mark.parametrize("forecaster_name, regressor, n_jobs_expected", 
    [('ForecasterDirect', LinearRegression(), 1),
     ('ForecasterDirect', HistGradientBoostingRegressor(), cpu_count() - 1),
     ('ForecasterDirect', LGBMRegressor(), 1),
     ('ForecasterDirect', LGBMRegressor(n_jobs=1), cpu_count() - 1),
     ('ForecasterDirectMultiVariate', LinearRegression(), 1),
     ('ForecasterDirectMultiVariate', HistGradientBoostingRegressor(), cpu_count() - 1),
     ('ForecasterDirectMultiVariate', LGBMRegressor(), 1),
     ('ForecasterDirectMultiVariate', LGBMRegressor(n_jobs=1), cpu_count() - 1),
     ('ForecasterRecursive', LinearRegression(), 1),
     ('ForecasterRecursive', HistGradientBoostingRegressor(), 1),
     ('ForecasterRecursive', LGBMRegressor(), 1),
     ('ForecasterRecursive', LGBMRegressor(n_jobs=1), 1)
], 
ids=lambda info: f'info: {info}')
def test_select_n_jobs_fit_forecaster(forecaster_name, regressor, n_jobs_expected):
    """
    Test select_n_jobs_fit_forecaster
    """
    n_jobs = select_n_jobs_fit_forecaster(
                 forecaster_name = forecaster_name, 
                 regressor       = regressor
             )
    
    assert n_jobs == n_jobs_expected



# Unit test create_predict_X ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import FunctionTransformer
from sklearn.compose import make_column_transformer, make_column_selector
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor
from lightgbm import LGBMRegressor

from ....exceptions import DataTransformationWarning
from skforecast.utils import transform_numpy
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect

# Fixtures
from .fixtures_forecaster_direct import y as y_categorical
from .fixtures_forecaster_direct import exog as exog_categorical
from .fixtures_forecaster_direct import data  # to test results when using differentiation


@pytest.mark.parametrize("steps", [[1, 2.0, 3], [1, 4.]], 
                         ids=lambda steps: f'steps: {steps}')
def test_create_predict_X_TypeError_when_steps_list_contain_floats(steps):
    """
    Test create_predict_X TypeError when steps is a list with floats.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(y=pd.Series(np.arange(10)))

    err_msg = re.escape(
        f"`steps` argument must be an int, a list of ints or `None`. "
        f"Got {type(steps)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        forecaster.create_predict_X(steps=steps)


def test_create_predict_X_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.create_predict_X(steps=5)


@pytest.mark.parametrize("steps", [3, [1, 2, 3], None], 
                         ids=lambda steps: f'steps: {steps}')
def test_create_predict_X_output(steps):
    """
    Test create_predict_X output.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=3)
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float)))
    results = forecaster.create_predict_X(steps=steps)

    expected = pd.DataFrame(
        data = {
            'lag_1': [49., 49., 49.],
            'lag_2': [48., 48., 48.],
            'lag_3': [47., 47., 47.]
        },
        index = pd.RangeIndex(start=50, stop=53, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_output_when_regressor_is_LinearRegression_with_list_interspersed():
    """
    Test create_predict_X output when steps is
    a list with interspersed steps.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float)))
    results = forecaster.create_predict_X(steps=[1, 4])

    expected = pd.DataFrame(
        data = {
            'lag_1': [49., 49.],
            'lag_2': [48., 48.],
            'lag_3': [47., 47.]
        },
        index = pd.Index([50, 53], dtype=int)
    )
    expected.index = expected.index.astype(results.index.dtype)
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_output_when_regressor_is_LinearRegression_using_last_window():
    """
    Test create_predict_X output when external last_window.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float)))
    last_window = pd.Series(
        data  = [47., 48., 49.], 
        index = pd.RangeIndex(start=47, stop=50, step=1)
    )
    results = forecaster.create_predict_X(steps=[1, 2, 3, 4], last_window=last_window)

    expected = pd.DataFrame(
        data = {
            'lag_1': [49., 49., 49., 49.],
            'lag_2': [48., 48., 48., 48.],
            'lag_3': [47., 47., 47., 47.]
        },
        index = pd.RangeIndex(start=50, stop=54, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_output_when_exog():
    """
    Test create_predict_X output when exog.
    """
    forecaster = ForecasterDirect(LinearRegression(), lags=3, steps=5)
    forecaster.fit(
        y    = pd.Series(np.arange(50, dtype=float)),
        exog = pd.Series(np.arange(start=100, stop=150, step=1, dtype=float), name="exog")
    )
    results = forecaster.create_predict_X(
                  steps = 5, 
                  exog  = pd.Series(np.arange(start=25, stop=50, step=0.5, dtype=float),
                                    index=pd.RangeIndex(start=50, stop=100),
                                    name="exog")
               )

    expected = pd.DataFrame(
        data = {
            'lag_1': [49., 49., 49., 49., 49.],
            'lag_2': [48., 48., 48., 48., 48.],
            'lag_3': [47., 47., 47., 47., 47.],
            'exog': [25., 25.5, 26., 26.5, 27.]
        },
        index = pd.RangeIndex(start=50, stop=55, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_output_with_transform_y():
    """
    Test create_predict_X output when StandardScaler.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    transformer_y = StandardScaler()

    forecaster = ForecasterDirect(
                     regressor     = LinearRegression(),
                     lags          = 5,
                     steps         = 5,
                     transformer_y = transformer_y,
                 )
    forecaster.fit(y=y)
    results = forecaster.create_predict_X()

    expected = pd.DataFrame(
        data = {
            'lag_1': [-0.52297655, -0.52297655, -0.52297655, -0.52297655, -0.52297655],
            'lag_2': [-0.28324197, -0.28324197, -0.28324197, -0.28324197, -0.28324197],
            'lag_3': [-0.45194408, -0.45194408, -0.45194408, -0.45194408, -0.45194408],
            'lag_4': [-0.30987914, -0.30987914, -0.30987914, -0.30987914, -0.30987914],
            'lag_5': [-0.1056608, -0.1056608, -0.1056608, -0.1056608, -0.1056608]
        },
        index = pd.RangeIndex(start=20, stop=25, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_create_predict_X_output_with_transform_y_and_transform_exog(n_jobs):
    """
    Test create_predict_X output when StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4 ,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8 ,
                      -0.61, -0.88])
        )
    exog = pd.DataFrame({
                'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4, 60.3, 87.2,
                          7.5, 60.4, 50.3, 57.3, 24.7, 87.4, 87.2, 60.4, 50.7, 7.5],
                'col_2': ['a'] * 10 + ['b'] * 10}
           )
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=20, stop=40)

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['col_1']),
                            ('onehot', OneHotEncoder(), ['col_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     lags             = 5,
                     steps            = 5,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog,
                     n_jobs           = n_jobs
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster.create_predict_X(steps=[1, 2, 3, 4, 5], exog=exog_predict)

    expected = pd.DataFrame(
        data = {
            'lag_1': [-0.52297655, -0.52297655, -0.52297655, -0.52297655, -0.52297655],
            'lag_2': [-0.28324197, -0.28324197, -0.28324197, -0.28324197, -0.28324197],
            'lag_3': [-0.45194408, -0.45194408, -0.45194408, -0.45194408, -0.45194408],
            'lag_4': [-0.30987914, -0.30987914, -0.30987914, -0.30987914, -0.30987914],
            'lag_5': [-0.1056608, -0.1056608, -0.1056608, -0.1056608, -0.1056608],
            'col_1': [-1.7093071, -1.0430105, 0.372377, 0.2540995, -0.006111],
            'col_2_a': [1., 1., 1., 1., 1.],
            'col_2_b': [0., 0., 0., 0., 0.]
        },
        index = pd.RangeIndex(start=20, stop=25, step=1)
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_output_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test create_predict_X output when using HistGradientBoostingRegressor 
    and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog_categorical,
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterDirect(
                     regressor        = HistGradientBoostingRegressor(
                                            categorical_features = categorical_features,
                                            random_state         = 123
                                        ),
                     lags             = 5,
                     steps            = 10, 
                     transformer_y    = None,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    results = forecaster.create_predict_X(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
        data = {
            'lag_1': [0.61289453, 0.61289453, 0.61289453, 0.61289453, 0.61289453,
                      0.61289453, 0.61289453, 0.61289453, 0.61289453, 0.61289453],
            'lag_2': [0.51948512, 0.51948512, 0.51948512, 0.51948512, 0.51948512,
                      0.51948512, 0.51948512, 0.51948512, 0.51948512, 0.51948512],
            'lag_3': [0.98555979, 0.98555979, 0.98555979, 0.98555979, 0.98555979,
                      0.98555979, 0.98555979, 0.98555979, 0.98555979, 0.98555979],
            'lag_4': [0.48303426, 0.48303426, 0.48303426, 0.48303426, 0.48303426,
                      0.48303426, 0.48303426, 0.48303426, 0.48303426, 0.48303426],
            'lag_5': [0.25045537, 0.25045537, 0.25045537, 0.25045537, 0.25045537,
                      0.25045537, 0.25045537, 0.25045537, 0.25045537, 0.25045537],
            'exog_2': [0., 1., 2., 3., 4., 0., 1., 2., 3., 4.],
            'exog_3': [0., 1., 2., 3., 4., 0., 1., 2., 3., 4.],
            'exog_1': [0.12062867, 0.8263408, 0.60306013, 0.54506801, 0.34276383,
                       0.30412079, 0.41702221, 0.68130077, 0.87545684, 0.51042234]
        },
        index = pd.RangeIndex(start=50, stop=60, step=1)
    ).astype({'exog_2': int, 'exog_3': int})
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_categorical_features_auto_detect_LGBMRegressor():
    """
    Test create_predict_X when using LGBMRegressor and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog_categorical,
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    pipeline_categorical = make_pipeline(
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               FunctionTransformer(
                                   func=lambda x: x.astype('category'),
                                   feature_names_out= 'one-to-one'
                               )
                           )

    transformer_exog = make_column_transformer(
                           (
                               pipeline_categorical,
                               make_column_selector(dtype_exclude=np.number)
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterDirect(
                     regressor        = LGBMRegressor(verbose=-1, random_state=123),
                     lags             = 5,
                     steps            = 10, 
                     transformer_y    = None,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    results = forecaster.create_predict_X(steps=10, exog=exog_predict)

    expected = pd.DataFrame(
        data = {
            'lag_1': [0.61289453, 0.61289453, 0.61289453, 0.61289453, 0.61289453,
                      0.61289453, 0.61289453, 0.61289453, 0.61289453, 0.61289453],
            'lag_2': [0.51948512, 0.51948512, 0.51948512, 0.51948512, 0.51948512,
                      0.51948512, 0.51948512, 0.51948512, 0.51948512, 0.51948512],
            'lag_3': [0.98555979, 0.98555979, 0.98555979, 0.98555979, 0.98555979,
                      0.98555979, 0.98555979, 0.98555979, 0.98555979, 0.98555979],
            'lag_4': [0.48303426, 0.48303426, 0.48303426, 0.48303426, 0.48303426,
                      0.48303426, 0.48303426, 0.48303426, 0.48303426, 0.48303426],
            'lag_5': [0.25045537, 0.25045537, 0.25045537, 0.25045537, 0.25045537,
                      0.25045537, 0.25045537, 0.25045537, 0.25045537, 0.25045537],
            'exog_2': [0., 1., 2., 3., 4., 0., 1., 2., 3., 4.],
            'exog_3': [0., 1., 2., 3., 4., 0., 1., 2., 3., 4.],
            'exog_1': [0.12062867, 0.8263408, 0.60306013, 0.54506801, 0.34276383,
                       0.30412079, 0.41702221, 0.68130077, 0.87545684, 0.51042234]
        },
        index = pd.RangeIndex(start=50, stop=60, step=1)
    ).astype({'exog_2': int, 'exog_3': int}
    ).astype({'exog_2': 'category', 'exog_3': 'category'})
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_regressor_is_LinearRegression_with_exog_differentiation_is_1_and_transformer_y_steps_1():
    """
    Test create_predict_X when using LinearRegression as regressor and differentiation=1,
    and transformer_y is StandardScaler, steps=1.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )

    forecaster = ForecasterDirect(
                     regressor       = LinearRegression(),
                     lags            = [1, 5],
                     steps           = 1,
                     differentiation = 1
                )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    results = forecaster.create_predict_X(exog=exog.loc[end_train:])

    expected = pd.DataFrame(
        data = np.array([
            [0.07503713, -0.01018012,  1.16172882]]
        ),
        columns = ['lag_1', 'lag_5', 'exog'],
        index = pd.date_range(start='2003-04-01', periods=1, freq='MS')
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_regressor_is_LinearRegression_with_exog_differentiation_is_1_and_transformer_y_steps_5():
    """
    Test create_predict_X when using LinearRegression as regressor and differentiation=1,
    and transformer_y is StandardScaler, steps=5.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )

    forecaster = ForecasterDirect(
                     regressor       = LinearRegression(),
                     lags            = [1, 5],
                     steps           = 5,
                     differentiation = 1
                )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    results = forecaster.create_predict_X(exog=exog.loc[end_train:])

    expected = pd.DataFrame(
        data = np.array([
            [0.07503713, -0.01018012,  1.16172882],
            [0.07503713, -0.01018012,  0.29468848],
            [0.07503713, -0.01018012, -0.4399757 ],
            [0.07503713, -0.01018012,  1.25008389],
            [0.07503713, -0.01018012,  1.37496887]]
        ),
        columns = ['lag_1', 'lag_5', 'exog'],
        index = pd.date_range(start='2003-04-01', periods=5, freq='MS')
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_window_features_steps_1():
    """
    Test the output of create_predict_X when using window_features and exog 
    with datetime index and steps=1.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    exog_datetime_pred = pd.Series(
        np.arange(115, 120), index=pd.date_range('2000-01-16', periods=5, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterDirect(
        LinearRegression(), steps=1, lags=5, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    results = forecaster.create_predict_X(exog=exog_datetime_pred)

    expected = pd.DataFrame(
        data = np.array([
                    [14., 13., 12., 11., 10., 12., 12., 69., 115.]]),
        index   = pd.date_range('2000-01-16', periods=1, freq='D'),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                   'roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog']
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_window_features_steps_10():
    """
    Test the output of create_predict_X when using window_features and exog 
    with datetime index and steps=10.
    """
    y_datetime = pd.Series(
        np.arange(20), index=pd.date_range('2000-01-01', periods=20, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.DataFrame(
        {'exog_1': np.arange(100, 120, dtype=float), 
         'exog_2': np.arange(200, 220, dtype=float)},
        index=pd.date_range('2000-01-01', periods=20, freq='D')
    )
    exog_datetime_pred = pd.DataFrame(
        {'exog_1': np.arange(120, 130, dtype=float),
         'exog_2': np.arange(220, 230, dtype=float)},
        index=pd.date_range('2000-01-21', periods=10, freq='D')
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterDirect(
        LinearRegression(), steps=10, lags=5, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    results = forecaster.create_predict_X(exog=exog_datetime_pred)

    expected = pd.DataFrame(
        data = np.array([
                    [19., 18., 17., 16., 15., 17., 17., 99., 120., 220.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 121., 221.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 122., 222.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 123., 223.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 124., 224.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 125., 225.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 126., 226.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 127., 227.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 128., 228.],
                    [19., 18., 17., 16., 15., 17., 17., 99., 129., 229.]]),
        index   = pd.date_range('2000-01-21', periods=10, freq='D'),
        columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                   'roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_when_window_features_and_lags_None_steps_10():
    """
    Test the output of create_predict_X when using window_features and exog 
    with datetime index, lags=None and steps=10.
    """
    y_datetime = pd.Series(
        np.arange(20), index=pd.date_range('2000-01-01', periods=20, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.DataFrame(
        {'exog_1': np.arange(100, 120, dtype=float), 
         'exog_2': np.arange(200, 220, dtype=float)},
        index=pd.date_range('2000-01-01', periods=20, freq='D')
    )
    exog_datetime_pred = pd.DataFrame(
        {'exog_1': np.arange(120, 130, dtype=float),
         'exog_2': np.arange(220, 230, dtype=float)},
        index=pd.date_range('2000-01-21', periods=10, freq='D')
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterDirect(
        LinearRegression(), steps=10, lags=None, window_features=[rolling, rolling_2]
    )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    results = forecaster.create_predict_X(exog=exog_datetime_pred)

    expected = pd.DataFrame(
        data = np.array([
                    [17., 17., 99., 120., 220.],
                    [17., 17., 99., 121., 221.],
                    [17., 17., 99., 122., 222.],
                    [17., 17., 99., 123., 223.],
                    [17., 17., 99., 124., 224.],
                    [17., 17., 99., 125., 225.],
                    [17., 17., 99., 126., 226.],
                    [17., 17., 99., 127., 227.],
                    [17., 17., 99., 128., 228.],
                    [17., 17., 99., 129., 229.]]),
        index   = pd.date_range('2000-01-21', periods=10, freq='D'),
        columns = ['roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog_1', 'exog_2']
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_predict_X_same_predictions_as_predict():
    """
    Test create_predict_X matrix returns the same predictions as predict method
    when passing to the regressor predict method.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     lags             = [1, 5],
                     window_features  = [rolling, rolling_2],
                     steps            = 6,
                     transformer_y    = None,
                     transformer_exog = None,
                     differentiation  = None
                 )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    X_predict = forecaster.create_predict_X(exog=exog.loc[end_train:])

    for i, step in enumerate(forecaster.steps):
        results = forecaster.regressors_[step].predict(X_predict.iloc[[i]])
        expected = forecaster.predict(steps=[step], exog=exog.loc[end_train:]).to_numpy()
        np.testing.assert_array_almost_equal(results, expected, decimal=7)


def test_create_predict_X_same_predictions_as_predict_transformers():
    """
    Test create_predict_X matrix returns the same predictions as predict method
    when passing to the regressor predict method with transformation.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     lags             = [1, 5],
                     window_features  = [rolling, rolling_2],
                     steps            = 6,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                     differentiation  = None
                 )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations or differentiation in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict = forecaster.create_predict_X(exog=exog.loc[end_train:])

    for i, step in enumerate(forecaster.steps):
        results = forecaster.regressors_[step].predict(X_predict.iloc[[i]])
        results = transform_numpy(
                      array             = results,
                      transformer       = forecaster.transformer_y,
                      fit               = False,
                      inverse_transform = True
                  )
        expected = forecaster.predict(steps=[step], exog=exog.loc[end_train:]).to_numpy()
        np.testing.assert_array_almost_equal(results, expected, decimal=7)


def test_create_predict_X_same_predictions_as_predict_transformers_diff():
    """
    Test create_predict_X matrix returns the same predictions as predict method
    when passing to the regressor predict method with transformation and differentiation.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=[5, 5])
    rolling_2 = RollingFeatures(stats='sum', window_sizes=[6])

    forecaster = ForecasterDirect(
                     regressor        = LinearRegression(),
                     lags             = [1, 5],
                     window_features  = [rolling, rolling_2],
                     steps            = 6,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                     differentiation  = 1
                 )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations or differentiation in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict = forecaster.create_predict_X(exog=exog.loc[end_train:])

    for i, step in enumerate(forecaster.steps):
        results = forecaster.regressors_[step].predict(X_predict.iloc[[i]])
        results = forecaster.differentiator.inverse_transform_next_window(results)
        results = transform_numpy(
                      array             = results,
                      transformer       = forecaster.transformer_y,
                      fit               = False,
                      inverse_transform = True
                  )
        expected = forecaster.predict(steps=[step], exog=exog.loc[end_train:]).to_numpy()
        np.testing.assert_array_almost_equal(results, expected, decimal=7)



# Unit test check_exog_dtypes
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.utils import check_exog_dtypes
from skforecast.exceptions import DataTypeWarning


@pytest.mark.parametrize("exog", 
                         [pd.Series(['A', 'B', 'C']),
                          pd.Series(['A', 'B', 'C']).to_frame()], 
                         ids = lambda exog: f'exog type: {type(exog)}')
def test_check_exog_dtypes_DataTypeWarning_when_exog_has_str_values(exog):
    """
    Check DataTypeWarning is issued when exog is pandas Series or DataFrame 
    with missing values.
    """
    warn_msg = re.escape(
        "`exog` may contain only `int`, `float` or `category` dtypes. Most "
        "machine learning models do not allow other types of values. "
        "Fitting the forecaster may fail."
    )
    with pytest.warns(DataTypeWarning, match = warn_msg):
        check_exog_dtypes(exog, call_check_exog=False)


@pytest.mark.parametrize("exog", 
                         [pd.Series([True, True, True], name='exog'),
                          pd.Series([True, True, True], name='exog').to_frame()], 
                         ids = lambda exog: f'exog type: {type(exog)}')
def test_check_exog_dtypes_DataTypeWarning_when_exog_has_bool_values(exog):
    """
    Check DataTypeWarning is issued when exog is pandas Series or DataFrame 
    with missing values.
    """
    warn_msg = re.escape(
        "`exog` may contain only `int`, `float` or `category` dtypes. Most "
        "machine learning models do not allow other types of values. "
        "Fitting the forecaster may fail."
    )
    with pytest.warns(DataTypeWarning, match = warn_msg):
        check_exog_dtypes(exog)


def test_check_exog_dtypes_TypeError_when_exog_is_Series_with_no_int_categories():
    """
    Check TypeError is raised when exog is pandas Series with no integer
    categories.
    """
    err_msg = re.escape(
        ("Categorical dtypes in exog must contain only integer values. "
         "See skforecast docs for more info about how to include "
         "categorical features https://skforecast.org/"
         "latest/user_guides/categorical-features.html")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_exog_dtypes(pd.Series(['A', 'B', 'C'], dtype='category', name='exog'))


def test_check_exog_dtypes_TypeError_when_exog_is_DataFrame_with_no_int_categories():
    """
    Check TypeError is raised when exog is pandas DataFrame with no integer
    categories.
    """
    err_msg = re.escape(
        ("Categorical dtypes in exog must contain only integer values. "
         "See skforecast docs for more info about how to include "
         "categorical features https://skforecast.org/"
         "latest/user_guides/categorical-features.html")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_exog_dtypes(pd.Series(['A', 'B', 'C'], dtype='category', name='exog').to_frame())


@pytest.mark.parametrize("dtype", 
                         ['int8', 'int16', 'int32', 'int64',
                          'float16', 'float32', 'float64',
                          'category'], 
                         ids = lambda exog: f'exog type: {type(exog)}')
def test_check_exog_dtypes_pandas_Series(dtype):
    """
    Test check_exog_dtypes accepts all dtypes in a pandas Series.
    """
    series = pd.Series(np.array([1, 2, 3]), name='exog', dtype=dtype) 
    _ = check_exog_dtypes(series)

    assert _ is None


def test_check_exog_dtypes_pandas_DataFrame():
    """
    Test check_exog_dtypes accepts all dtypes in a pandas DataFrame.
    """
    df = pd.DataFrame({'col1': [1., 2., 3.], 'col2': [4, 5, 6]}) 
    df['col2'] = df['col2'].astype('category')
    _ = check_exog_dtypes(df)

    assert _ is None



# Unit test check_preprocess_series
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.exceptions import IgnoredArgumentWarning, InputTypeWarning
from skforecast.preprocessing import reshape_series_wide_to_long
from skforecast.utils import check_preprocess_series
from skforecast.recursive.tests.tests_forecaster_recursive_multiseries.fixtures_forecaster_recursive_multiseries import (
    series_wide_range,
    series_wide_dt,
    series_long_dt,
    series_dict_range,
    series_dict_dt
) 


def test_TypeError_check_preprocess_series_when_series_is_not_pandas_DataFrame_or_dict():
    """
    Test TypeError is raised when series is not a pandas DataFrame or dict.
    """
    series = np.array([1, 2, 3])

    err_msg = re.escape(
        f"`series` must be a pandas DataFrame or a dict of DataFrames or Series. "
        f"Got {type(series)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_series(series = series)


def test_TypeError_check_preprocess_series_when_series_is_pandas_DataFrame_MultiIndex_without_datetime():
    """
    Test check_preprocess_series when `series` is a pandas DataFrame with 
    a MultiIndex without datetime in the second level.
    """
    series = pd.DataFrame({
        '1': pd.Series(np.arange(6), index=pd.MultiIndex.from_product([['a', 'b'], range(3)])),
    })

    err_msg = re.escape(
        "The second level of the MultiIndex in `series` must be a pandas "
        "DatetimeIndex with the same frequency for each series. "
        "Found <class 'pandas.core.indexes.base.Index'>."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_series(series=series)


def test_TypeError_check_preprocess_series_when_series_is_dict_with_no_pandas_Series_or_DataFrame():
    """
    Test TypeError is raised when series is dict not containing 
    a pandas Series or DataFrame.
    """
    series_dict = {'l1': np.array([1, 2, 3]), 'l2': pd.Series([1, 2, 3])}
    err_msg = re.escape(
        "If `series` is a dictionary, all series must be a named pandas Series "
        "or a pandas DataFrame with a single column. Review series: ['l1']"
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_ValueError_check_preprocess_series_when_series_is_dict_with_a_DataFrame_with_2_columns():
    """
    Test ValueError is raised when series is dict containing a pandas DataFrame 
    with 2 columns.
    """
    series_dict = {'l1': series_wide_dt, 'l2': series_wide_dt}
    err_msg = re.escape(
        "If `series` is a dictionary, all series must be a named pandas Series "
        "or a pandas DataFrame with a single column. Review series: 'l1'"
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_TypeError_check_preprocess_series_when_series_is_dict_with_no_DatetimeIndex_RangeIndex():
    """
    Test TypeError is raised when series is dict containing series with no
    DatetimeIndex.
    """
    series_dict = {
        'l1': pd.Series(np.arange(3), index=pd.Index([0, 1, 2])),
        'l2': pd.Series(np.arange(3), index=pd.RangeIndex(start=0, stop=3, step=1))
    }

    err_msg = re.escape(
        "If `series` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "Review series: ['l1']"
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_ValueError_check_preprocess_series_when_series_is_dict_with_different_freqs_or_None():
    """
    Test ValueError is raised when series is dict containing series withou frequency 
    or with DatetimeIndex but different frequencies.
    """
    series_dict = {
        'l1': pd.Series(np.arange(10)),
        'l2': pd.Series(np.arange(10))
    }
    series_dict['l1'].index = pd.date_range(start='2000-01-01', periods=10, freq='D')
    series_dict['l2'].index = pd.date_range(start='2000-01-01', periods=10, freq='D')
    series_dict['l1'].index.freq = None

    err_msg = re.escape(
        "If `series` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "If it a MultiIndex DataFrame, the second level must be a DatetimeIndex "
        "with the same frequency for each series. Found series with no "
        "frequency or step."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series = series_dict)
    
    series_dict = {
        'l1': pd.Series(np.arange(10)),
        'l2': pd.Series(np.arange(10))
    }
    series_dict['l1'].index = pd.date_range(start='2000-01-01', periods=10, freq='D')
    series_dict['l2'].index = pd.date_range(start='2000-01-01', periods=10, freq='2D')

    err_msg = re.escape(
        "If `series` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "If it a MultiIndex DataFrame, the second level must be a DatetimeIndex "
        "with the same frequency for each series. "
        "Found frequencies: ['2D', 'D']"
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_ValueError_check_preprocess_series_when_series_is_DataFrame_MultiIndex_different_freqs_or_None():
    """
    Test ValueError is raised when series is a pandas DataFrame with a MultiIndex
    containing at least one series without frequency or with different frequencies.
    """
    series_wide = pd.DataFrame(
        {
            'series_id': ['1', '1', '1', '1', '2', '2', '2', '2'],
            'datetime': pd.date_range(start='2000-01-01', periods=8, freq='D'),
            'value': np.random.rand(8)
        }
    )
    series = series_wide.groupby("series_id", sort=False).apply(
        lambda x: x.set_index("datetime").asfreq("D"), include_groups=False
    )
    series = series.drop(index=('1', '2000-01-03'))

    err_msg = re.escape(
        "If `series` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "If it a MultiIndex DataFrame, the second level must be a DatetimeIndex "
        "with the same frequency for each series. Found series with no "
        "frequency or step."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series=series)
    
    series_wide = pd.DataFrame({
            'series_id': ['1', '1', '1', '1', '2', '2', '2', '2'],
            'datetime': pd.date_range(start='2000-01-01', periods=8, freq='D'),
            'value': np.random.rand(8)
        }
    )
    series = series_wide.groupby("series_id", sort=False).apply(
        lambda x: x.set_index("datetime").asfreq("D"), include_groups=False
    )
    series = series.drop(index=('1', '2000-01-02'))
    series = series.drop(index=('1', '2000-01-03'))

    err_msg = re.escape(
        "If `series` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "If it a MultiIndex DataFrame, the second level must be a DatetimeIndex "
        "with the same frequency for each series. "
        "Found frequencies: ['3D', 'D']"
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series=series)


def test_ValueError_check_preprocess_series_when_all_values_in_series_dict_are_nan():
    """
    Test ValueError is raised when all values in a series dict are NaN.
    """
    series = {
        '1': pd.Series(np.arange(7), index = pd.date_range(start='2022-01-01', periods=7, freq='D')),
        '2': pd.Series([np.nan] * 7, index = pd.date_range(start='2022-01-01', periods=7, freq='D'))
    }

    err_msg = re.escape("All values of series '2' are NaN.")
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series=series)
    

def test_ValueError_check_preprocess_series_when_all_series_values_are_missing_DataFrame_MultiIndex():
    """
    Test ValueError is raised when all series values are missing when series
    is a dict.
    """
    series_nan = pd.DataFrame(
        {'l1': pd.Series(np.arange(7)), 
         'l2': pd.Series([np.nan] * 7)}
    )
    series_nan.index = pd.date_range(start='2022-01-01', periods=7, freq='D')
    series_long_nan = reshape_series_wide_to_long(series_nan)

    err_msg = re.escape("All values of series 'l2' are NaN.")
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series=series_long_nan)


@pytest.mark.parametrize("series", 
                         [series_wide_dt, series_long_dt],
                         ids = lambda series: f'series index type: {type(series.index)}')
def test_check_preprocess_series_when_series_is_pandas_DataFrame_wide_or_MultiIndex_datetime(series):
    """
    Test check_preprocess_series when `series` is a pandas DataFrame wide or 
    long format.
    """
    warn_msg = re.escape(
        "Passing a DataFrame (either wide or long format) as `series` requires "
        "additional internal transformations, which can increase computational "
        "time. It is recommended to use a dictionary of pandas Series instead. "
        "For more details, see: "
        "https://skforecast.org/latest/user_guides/independent-multi-time-series-forecasting.html#input-data"
    )
    with pytest.warns(InputTypeWarning, match=warn_msg):
        series_dict, series_indexes = check_preprocess_series(series=series)

    expected_series_dict = {
        '1': pd.Series(
                 data  = series_dict_dt['l1'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=50, freq='D'),
                 name  ='1'
             ),
        '2': pd.Series(
                 data  = series_dict_dt['l2'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=50, freq='D'),
                 name  ='2'
             ),
    }

    expected_series_indexes = {
        col: pd.date_range(start='2000-01-01', periods=len(series_long_dt.loc[col]), freq='D')
        for col in series_long_dt.index.levels[0]
    }

    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k, v in series_dict.items():
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert k == v.name

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['1', '2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].freq == expected_series_indexes[k].freq


def test_check_preprocess_series_when_series_is_pandas_DataFrame_MultiIndex_datetime_with_two_columns():
    """
    Test check_preprocess_series when `series` is a pandas DataFrame with a MultiIndex
    containing two columns.
    """
    series_long_dt_two_columns = series_long_dt.copy()
    series_long_dt_two_columns['value_2'] = series_long_dt_two_columns['value']

    warn_msg = re.escape(
        "`series` DataFrame has multiple columns. Only the values of "
        "first column, 'value', will be used as series values. "
        "All other columns will be ignored."
    )
    with pytest.warns(IgnoredArgumentWarning, match=warn_msg):
        series_dict, series_indexes = check_preprocess_series(series=series_long_dt_two_columns)

    expected_series_dict = {
        '1': pd.Series(
                 data  = series_dict_dt['l1'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=50, freq='D'),
                 name  ='1'
             ),
        '2': pd.Series(
                 data  = series_dict_dt['l2'].to_numpy(),
                 index = pd.date_range(start='2000-01-01', periods=50, freq='D'),
                 name  ='2'
             ),
    }

    expected_series_indexes = {
        col: pd.date_range(start='2000-01-01', periods=len(series_long_dt_two_columns.loc[col]), freq='D')
        for col in series_long_dt_two_columns.index.levels[0]
    }

    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k, v in series_dict.items():
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert k == v.name

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['1', '2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].freq == expected_series_indexes[k].freq


@pytest.mark.parametrize("series", 
                         [series_wide_range, series_dict_range],
                         ids = lambda series: f'series type: {type(series)}')
def test_check_preprocess_series_when_series_is_DataFrame_or_dict_with_RangeIndex(series):
    """
    Test check_preprocess_series when `series` is a dict or a pandas DataFrame with
    a RangeIndex.
    """
    if isinstance(series, pd.DataFrame):
        series = series.rename(columns={'1': 'l1', '2': 'l2'})
    
    series_dict, series_indexes = check_preprocess_series(series=series)

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_dict_range['l1'].to_numpy(),
                  index = pd.RangeIndex(start=0, stop=len(series_dict_range['l1']), step=1),
                  name  ='l1'
              ),
        'l2': pd.Series(
                  data  = series_dict_range['l2'].to_numpy(),
                  index = pd.RangeIndex(start=0, stop=len(series_dict_range['l2']), step=1),
                  name  ='l2'
              ),
    }

    expected_series_indexes = {
        k: v.index
        for k, v in expected_series_dict.items()
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert k == series_dict[k].name

    indexes_step = [f'{v.index.step}' for v in series_dict.values()]
    assert len(set(indexes_step)) == 1

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['l1', 'l2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].step == expected_series_indexes[k].step


def test_check_preprocess_series_when_series_is_dict_with_a_pandas_Series_and_a_DataFrame():
    """
    Test check_preprocess_series when `series` is a dict containing a pandas Series
    and a pandas DataFrame with a single column.
    """
    series_dict = series_dict_dt.copy()
    series_dict['l2'] = series_dict['l2'].to_frame()

    series_dict, series_indexes = check_preprocess_series(series=series_dict)

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_dict['l1'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_dict['l1']), freq='D'),
                  name  ='l1'
              ),
        'l2': pd.Series(
                  data  = series_dict['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series_dict['l2']), freq='D'),
                  name  ='l2'
              ),
    }

    expected_series_indexes = {
        k: v.index
        for k, v in expected_series_dict.items()
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert k == series_dict[k].name

    indexes_freq = [f'{v.index.freq}' for v in series_dict.values()]
    assert len(set(indexes_freq)) == 1

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['l1', 'l2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].freq == expected_series_indexes[k].freq



# Unit test __init__ ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from ....exceptions import DataTransformationWarning
from ....preprocessing import RollingFeatures, TimeSeriesDifferentiator
from ....recursive import ForecasterRecursiveMultiSeries


def test_init_ValueError_when_no_lags_or_window_features():
    """
    Test ValueError is raised when no lags or window_features are passed.
    """
    err_msg = re.escape(
        "At least one of the arguments `lags` or `window_features` "
        "must be different from None. This is required to create the "
        "predictors used in training the forecaster."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor       = LinearRegression(),
            lags            = None,
            window_features = None
        )


@pytest.mark.parametrize("lags, window_features, expected", 
                         [(5, None, 5), 
                          (None, True, 6), 
                          ([], True, 6), 
                          (5, True, 6)], 
                         ids = lambda dt: f'lags, window_features, expected: {dt}')
def test_init_window_size_correctly_stored(lags, window_features, expected):
    """
    Test window_size is correctly stored when lags or window_features are passed.
    """
    if window_features:
        window_features = RollingFeatures(
            stats=['ratio_min_max', 'median'], window_sizes=[5, 6]
        )

    forecaster = ForecasterRecursiveMultiSeries(
                     regressor       = LinearRegression(),
                     lags            = lags,
                     window_features = window_features
                 )
    
    assert forecaster.window_size == expected
    if lags:
        np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
        assert forecaster.lags_names == [f'lag_{i}' for i in range(1, lags + 1)]
        assert forecaster.max_lag == lags
    else:
        assert forecaster.lags is None
        assert forecaster.lags_names is None
        assert forecaster.max_lag is None
    if window_features:
        assert forecaster.window_features_names == ['roll_ratio_min_max_5', 'roll_median_6']
        assert forecaster.window_features_class_names == ['RollingFeatures']
    else:
        assert forecaster.window_features_names is None
        assert forecaster.window_features_class_names is None


def test_init_ValueError_invalid_encoding():
    """
    Test ValueError is raised when encoding is not valid.
    """

    err_msg = re.escape(
        "Argument `encoding` must be one of the following values: 'ordinal', "
        "'ordinal_category', 'onehot' or None. Got 'invalid_encoding'."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor = LinearRegression(),
            lags      = [1, 2, 3],
            encoding  = 'invalid_encoding',
        )


def test_ForecasterRecursiveMultiSeries_init_not_scaling_with_linear_model():
    """
    Test Warning is raised when Forecaster has no transformer_series and it
    is using a linear model.
    """

    warn_msg = re.escape(
        "When using a linear model, it is recommended to use a transformer_series "
        "to ensure all series are in the same scale. You can use, for example, a "
        "`StandardScaler` from sklearn.preprocessing."
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        ForecasterRecursiveMultiSeries(
            regressor = LinearRegression(),
            lags      = [1, 2, 3]
        )


def test_init_TypeError_transformer_series_dict_encoding_None():
    """
    Test TypeError is raised when transformer_series is a dictionary 
    and encoding is None.
    """

    err_msg = re.escape(
        "When `encoding` is None, `transformer_series` must be a single "
        "transformer (not `dict`) as it is applied to all series."
    )
    with pytest.raises(TypeError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor          = LinearRegression(),
            lags               = [1, 2, 3],
            encoding           = None,
            transformer_series = {'1': StandardScaler(), '_unknown_level': StandardScaler()}
        )


def test_init_ValueError_transformer_series_dict_with_no_unknown_level():
    """
    Test ValueError is raised when transformer_series is a dictionary 
    and no '_unknown_level' key is provided.
    """

    err_msg = re.escape(
        "If `transformer_series` is a `dict`, a transformer must be "
        "provided to transform series that do not exist during training. "
        "Add the key '_unknown_level' to `transformer_series`. "
        "For example: {'_unknown_level': your_transformer}."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor          = LinearRegression(),
            lags               = [1, 2, 3],
            encoding           = 'ordinal',
            transformer_series = {'1': StandardScaler()}
        )


@pytest.mark.parametrize("diff", 
                         [1.5, 'not_int_dict'], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_init_TypeError_when_differentiation_argument_is_not_int_or_dict(diff):
    """
    Test TypeError is raised when differentiation is not an int or dict.
    """
    err_msg = re.escape(
        f"When including `differentiation`, this argument must be "
        f"an integer (equal to or greater than 1) or a dict of "
        f"integers. Got {type(diff)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor       = LinearRegression(),
            lags            = 5,
            differentiation = diff
        )


@pytest.mark.parametrize("diff", 
                         [-1, 0], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_init_ValueError_when_differentiation_is_int_but_not_greater_than_0(diff):
    """
    Test ValueError is raised when differentiation is an int not greater than 0.
    """
    err_msg = re.escape(
        f"If `differentiation` is an integer, it must be equal "
        f"to or greater than 1. Got {diff}."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor       = LinearRegression(),
            lags            = 5,
            differentiation = diff
        )


@pytest.mark.parametrize("diff", 
                         [1, 2], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_init_when_differentiation_is_integer(diff):
    """
    Test differentiation is correctly stored when it is an integer.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor       = LinearRegression(),
                     lags            = 5,
                     differentiation = diff
                 )
    
    assert forecaster.differentiation == diff
    assert forecaster.differentiation_max == diff
    assert forecaster.window_size == 5 + diff
    assert isinstance(forecaster.differentiator, TimeSeriesDifferentiator)


def test_init_TypeError_when_differentiation_is_dict_with_encoding_None():
    """
    Test TypeError is raised when differentiation is a dict but encoding is None.
    """
    err_msg = re.escape(
        "When `encoding` is None, `differentiation` must be an "
        "integer equal to or greater than 1. Same differentiation "
        "must be applied to all series."
    )
    with pytest.raises(TypeError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor       = LinearRegression(),
            lags            = 5,
            encoding        = None,
            differentiation = {'l1': 1, 'l2': 1, '_unknown_level': 1}
        )


def test_init_ValueError_when_differentiation_is_dict_with_no_unknown_level():
    """
    Test ValueError is raised when differentiation is a dict but encoding is None.
    """
    err_msg = re.escape(
        "If `differentiation` is a `dict`, an order must be provided "
        "to differentiate series that do not exist during training. "
        "Add the key '_unknown_level' to `differentiation`. "
        "For example: {'_unknown_level': 1}."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor       = LinearRegression(),
            lags            = 5,
            encoding        = 'ordinal',
            differentiation = {'l1': 1, 'l2': 1}
        )


@pytest.mark.parametrize("diff", 
                         [{'l1': 1, 'l2': 'not_int', '_unknown_level': 1}, 
                          {'l1': None, 'l2': 0, '_unknown_level': 1}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_init_ValueError_when_differentiation_is_dict_with_int_not_greater_than_0(diff):
    """
    Test ValueError is raised when differentiation is a dict containing 
    integers not greater than 0 or not None.
    """
    err_msg = re.escape(
        f"If `differentiation` is a dict, the values must be "
        f"None or integers equal to or greater than 1. "
        f"Got {diff['l2']} for series 'l2'."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor       = LinearRegression(),
            lags            = 5,
            differentiation = diff
        )


def test_init_ValueError_when_differentiation_is_dict_with_all_None():
    """
    Test ValueError is raised when differentiation is a dict with all values None.
    """
    err_msg = re.escape(
        "If `differentiation` is a dict, at least one value must be "
        "different from None. Got all values equal to None. If you "
        "do not want to differentiate any series, set `differentiation` "
        "to None."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveMultiSeries(
            regressor       = LinearRegression(),
            lags            = 5,
            differentiation = {'l1': None, 'l2': None, '_unknown_level': None}
        )


@pytest.mark.parametrize("diff", 
                         [{'l1': 1, 'l2': 2, '_unknown_level': 1}, 
                          {'l1': None, 'l2': 1, '_unknown_level': 1}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_init_when_differentiation_is_dict(diff):
    """
    Test differentiation is correctly stored when it is a dict.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     regressor       = LinearRegression(),
                     lags            = 5,
                     differentiation = diff
                 )
    
    if diff['l1']:
        assert forecaster.differentiation == {'l1': 1, 'l2': 2, '_unknown_level': 1}
        assert forecaster.differentiation_max == 2
        assert forecaster.window_size == 5 + 2
        assert isinstance(forecaster.differentiator['l1'], TimeSeriesDifferentiator)
        assert isinstance(forecaster.differentiator['l2'], TimeSeriesDifferentiator)
        assert isinstance(forecaster.differentiator['_unknown_level'], TimeSeriesDifferentiator)
    else:
        assert forecaster.differentiation == {'l1': None, 'l2': 1, '_unknown_level': 1}
        assert forecaster.differentiation_max == 1
        assert forecaster.window_size == 5 + 1
        assert forecaster.differentiator['l1'] is None
        assert isinstance(forecaster.differentiator['l2'], TimeSeriesDifferentiator)
        assert isinstance(forecaster.differentiator['_unknown_level'], TimeSeriesDifferentiator)



# Unit test predict interval method with PyTorch backend
# ==============================================================================
import os
import re
import pytest
import numpy as np
import pandas as pd
os.environ["KERAS_BACKEND"] = "torch"
import keras
from skforecast.deep_learning.utils import create_and_compile_model
from skforecast.deep_learning import ForecasterRnn

series = pd.DataFrame(
    {
        "1": np.arange(50),
        "2": np.arange(50),
        "3": np.arange(50),
    },
    index=pd.date_range("2020-01-01", periods=50, freq="D")
)

exog = pd.DataFrame(
    {
        "exog1": np.arange(50),
        "exog2": np.arange(50),
    },
    index=pd.date_range("2020-01-01", periods=50, freq="D")
)

exog_pred = pd.DataFrame(
    {
        "exog1": np.arange(50, 60),
        "exog2": np.arange(50, 60),
    },
    index=pd.date_range("2020-02-20", periods=10, freq="D")
)

model = create_and_compile_model(
            series=series, 
            levels=["1", "2"],    
            lags=3,           
            steps=4,              
            recurrent_layer="LSTM",
            recurrent_units=128,
            dense_units=64,
        )

model_exog = create_and_compile_model(
            series=series, 
            exog=exog,
            levels=["1", "2", "3"],    
            lags=10,           
            steps=8,              
            recurrent_layer="LSTM",
            recurrent_units=128,
            dense_units=64,
        )


def test_predict_interval_ValueError_when_method_not_valid():
    """
    Test ValueError is raised when an invalid method is passed to predict_interval.
    """
    forecaster = ForecasterRnn(model, levels=["1", "2"], lags=3)
    forecaster.fit(series, store_in_sample_residuals=True)

    err_msg = re.escape(
        "Invalid `method` 'not_conformal'. Only 'conformal' is available."
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster.predict_interval(method="not_conformal")


def test_predict_interval_output_size_with_steps_by_default():
    """
    Test output sizes for predicting steps defined by default with intervals
    """
    # Create a ForecasterRnn object
    forecaster = ForecasterRnn(model, levels=["1", "2"], lags=3)
    forecaster.fit(series, store_in_sample_residuals=True)

    # Call the predict method
    int_preds = forecaster.predict_interval()

    # Check the shape and values of the predictions
    assert int_preds.shape == (4 * len(["1", "2"]), 4)


def test_predict_interval_output_size_3_steps_ahead():
    """
    Test output sizes for predicting 3 steps ahead with intervals
    """
    # Create a ForecasterRnn object
    forecaster = ForecasterRnn(model, levels=["1", "2"], lags=3)
    forecaster.fit(series, store_in_sample_residuals=True)

    # Call the predict method
    int_preds = forecaster.predict_interval(steps=3, interval=0.9)

    # Check the shape and values of the predictions
    assert int_preds.shape == (3 * len(["1", "2"]), 4)


def test_predict_interval_output_size_2_steps_ahead_specific_levels():
    """
    Test output sizes for predicting 2 steps ahead with intervals and specific levels
    """
    # Create a ForecasterRnn object
    forecaster = ForecasterRnn(model, levels=["1", "2"], lags=3)
    forecaster.fit(series, store_in_sample_residuals=True)

    # Call the predict method
    int_preds = forecaster.predict_interval(steps=2, levels="1")

    # Check the shape and values of the predictions
    assert int_preds.shape == (2 * 1, 4)


def test_predict_interval_exog_and_out_sample_residuals():
    """
    Test case for predicting with exogenous variables
    """
    forecaster = ForecasterRnn(model_exog, levels=["1", "2", "3"], lags=10)
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_

    predictions = forecaster.predict_interval(
        steps=None, exog=exog_pred, use_in_sample_residuals=False
    )

    assert predictions.shape == (24, 4)


def test_predict_interval_specific_levels_with_exog():
    """
    Test case for predicting with specific levels
    """
    forecaster = ForecasterRnn(model_exog, levels=["1", "2", "3"], lags=10)
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    predictions = forecaster.predict_interval(steps=5, exog=exog_pred, levels=["1", "2"])

    assert predictions.shape == (10, 4)



# Unit test _create_predict_inputs ForecasterSarimax
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.sarimax import Sarimax
from skforecast.recursive import ForecasterSarimax
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder

# Fixtures
from .fixtures_forecaster_sarimax import y
from .fixtures_forecaster_sarimax import y_lw
from .fixtures_forecaster_sarimax import exog
from .fixtures_forecaster_sarimax import exog_lw
from .fixtures_forecaster_sarimax import exog_predict
from .fixtures_forecaster_sarimax import exog_lw_predict
from .fixtures_forecaster_sarimax import y_datetime
from .fixtures_forecaster_sarimax import y_lw_datetime
from .fixtures_forecaster_sarimax import exog_datetime
from .fixtures_forecaster_sarimax import exog_lw_datetime
from .fixtures_forecaster_sarimax import exog_predict_datetime
from .fixtures_forecaster_sarimax import exog_lw_predict_datetime
from .fixtures_forecaster_sarimax import df_exog
from .fixtures_forecaster_sarimax import df_exog_lw
from .fixtures_forecaster_sarimax import df_exog_predict
from .fixtures_forecaster_sarimax import df_exog_lw_predict
from .fixtures_forecaster_sarimax import df_exog_datetime
from .fixtures_forecaster_sarimax import df_exog_lw_datetime
from .fixtures_forecaster_sarimax import df_exog_predict_datetime
from .fixtures_forecaster_sarimax import df_exog_lw_predict_datetime


def test_create_predict_inputs_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 1, 1)))

    err_msg = re.escape(
        ("This Forecaster instance is not fitted yet. Call `fit` with "
         "appropriate arguments before using predict.")
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster._create_predict_inputs(steps=5)


def test_create_predict_inputs_ValueError_when_ForecasterSarimax_last_window_exog_is_not_None_and_last_window_is_not_provided():
    """
    Check ValueError is raised when last_window_exog is not None, but 
    last_window is not provided.
    """
    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y, exog=exog)
    
    err_msg = re.escape(
        ("To make predictions unrelated to the original data, both "
         "`last_window` and `last_window_exog` must be provided.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_predict_inputs(
            steps            = 5, 
            exog             = exog_predict, 
            last_window      = None, 
            last_window_exog = exog
        )


def test_create_predict_inputs_ValueError_when_ForecasterSarimax_last_window_exog_is_None_and_included_exog_is_true():
    """
    Check ValueError is raised when last_window_exog is None, but included_exog
    is True and last_window is provided.
    """
    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y, exog=exog)
    
    err_msg = re.escape(
        ("Forecaster trained with exogenous variable/s. To make predictions "
         "unrelated to the original data, same variable/s must be provided "
         "using `last_window_exog`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_predict_inputs(
            steps            = 5, 
            exog             = exog_lw_predict, 
            last_window      = y_lw, 
            last_window_exog = None
        )


def test_create_predict_inputs_output_ForecasterSarimax():
    """
    Test _create_predict_inputs output of ForecasterSarimax.
    """
    forecaster = ForecasterSarimax(regressor = Sarimax())
    forecaster.fit(y=y)
    results = forecaster._create_predict_inputs(steps=5)

    expected = (
        None,
        None,
        None
    )
    
    assert results == expected


def test_create_predict_inputs_output_ForecasterSarimax_with_exog():
    """
    Test _create_predict_inputs output of ForecasterSarimax with exogenous variables.
    """
    forecaster = ForecasterSarimax(regressor = Sarimax())
    forecaster.fit(y=y, exog=exog)
    results = forecaster._create_predict_inputs(steps=5, exog=exog_predict)

    expected = (
        None,
        None,
        pd.DataFrame(
            {'exog': np.array([1.27501789, 1.31081724, 1.34284965, 1.37614005, 1.41412559])},
            index = pd.RangeIndex(start=50, stop=55, step=1)
        )
    )
    
    assert results[0] == expected[0]
    assert results[1] == expected[1]
    pd.testing.assert_frame_equal(results[2], expected[2])


def test_create_predict_inputs_output_ForecasterSarimax_with_transform_y():
    """
    Test _create_predict_inputs output of ForecasterSarimax with a 
    StandardScaler() as transformer_y.
    """        
    forecaster = ForecasterSarimax(
                     regressor     = Sarimax(),
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=y)
    results = forecaster._create_predict_inputs(steps=5)

    expected = (
        None,
        None,
        None
    )
    
    assert results == expected


def test_create_predict_inputs_output_ForecasterSarimax_with_transform_y_and_transform_exog():
    """
    Test _create_predict_inputs output of ForecasterSarimax, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """    
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterSarimax(
                     regressor        = Sarimax(),
                     transformer_y    = StandardScaler(),
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=df_exog)
    results = forecaster._create_predict_inputs(steps=5, exog=df_exog_predict)

    expected = (
        None,
        None,
        pd.DataFrame(
            {'exog_1': np.array([-0.20710692, 0.10403839, 0.38244384, 0.67178294, 1.00192923]),
             'exog_2_a': np.array([1., 0., 1., 0., 1.]),
             'exog_2_b': np.array([0., 1., 0., 1., 0.])},
            index = pd.RangeIndex(start=50, stop=55, step=1)
        )
    )
    
    assert results[0] == expected[0]
    assert results[1] == expected[1]
    pd.testing.assert_frame_equal(results[2], expected[2])


def test_create_predict_inputs_ValueError_when_last_window_index_does_not_follow_training_set():
    """
    Raise ValueError if `last_window` index does not start at the end 
    of the index seen by the forecaster.
    """
    y_test = pd.Series(data=y_datetime.to_numpy())
    y_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    lw_test = pd.Series(data=y_lw_datetime.to_numpy())
    lw_test.index = pd.date_range(start='2022-03-01', periods=50, freq='D')

    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 0, 1)))
    forecaster.fit(y=y_test)

    err_msg = re.escape(
        (f"To make predictions unrelated to the original data, `last_window` "
         f"has to start at the end of the index seen by the forecaster.\n"
         f"    Series last index         : 2022-02-19 00:00:00.\n"
         f"    Expected index            : 2022-02-20 00:00:00.\n"
         f"    `last_window` index start : 2022-03-01 00:00:00.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_predict_inputs(steps=5, last_window=lw_test)


def test_create_predict_inputs_ValueError_when_last_window_exog_index_does_not_follow_training_set():
    """
    Raise ValueError if `last_window_exog` index does not start at the end 
    of the index seen by the forecaster.
    """
    y_test = pd.Series(data=y_datetime.to_numpy())
    y_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    lw_test = pd.Series(data=y_lw_datetime.to_numpy())
    lw_test.index = pd.date_range(start='2022-02-20', periods=50, freq='D')
    
    exog_test = pd.Series(data=exog_datetime.to_numpy(), name='exog')
    exog_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D', name='exog')
    exog_pred_test = pd.Series(data=exog_predict_datetime.to_numpy(), name='exog')
    exog_pred_test.index = pd.date_range(start='2022-04-11', periods=10, freq='D', name='exog')
    lw_exog_test = pd.Series(data=exog_lw_datetime.to_numpy(), name='exog')
    lw_exog_test.index = pd.date_range(start='2022-03-01', periods=50, freq='D', name='exog')

    forecaster = ForecasterSarimax(regressor=Sarimax(order=(1, 0, 1)))
    forecaster.fit(y=y_test, exog=exog_test)

    err_msg = re.escape(
        (f"To make predictions unrelated to the original data, `last_window_exog` "
         f"has to start at the end of the index seen by the forecaster.\n"
         f"    Series last index              : 2022-02-19 00:00:00.\n"
         f"    Expected index                 : 2022-02-20 00:00:00.\n"
         f"    `last_window_exog` index start : 2022-03-01 00:00:00.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_predict_inputs(
            steps            = 5, 
            exog             = exog_pred_test, 
            last_window      = lw_test,
            last_window_exog = lw_exog_test
        )


def test_create_predict_inputs_output_ForecasterSarimax_with_last_window():
    """
    Test _create_predict_inputs output of ForecasterSarimax with `last_window`.
    """    
    forecaster = ForecasterSarimax(regressor = Sarimax())
    forecaster.fit(y=y_datetime)
    results = forecaster._create_predict_inputs(steps=5, last_window=y_lw_datetime)

    expected = (
        pd.Series(
            np.array([0.59418877, 0.70775844, 0.71950195, 0.74432369, 0.80485511,
                      0.78854235, 0.9710894 , 0.84683354, 0.46382252, 0.48527317,
                      0.5280586 , 0.56233647, 0.5885704 , 0.66948036, 0.67799365,
                      0.76299549, 0.79972374, 0.77052192, 0.99438934, 0.80054443,
                      0.49055721, 0.52440799, 0.53664948, 0.55209054, 0.60336564,
                      0.68124538, 0.67807535, 0.79489265, 0.7846239 , 0.8130087 ,
                      0.9777323 , 0.89308148, 0.51269597, 0.65299589, 0.5739764 ,
                      0.63923842, 0.70387188, 0.77064824, 0.84618588, 0.89272889,
                      0.89789988, 0.94728069, 1.05070727, 0.96965567, 0.57329151,
                      0.61850684, 0.61899573, 0.66520922, 0.72652015, 0.85586494]),
            index = pd.date_range(start='2050', periods=50, freq='YE'),
            name = 'y'
        ),
        None,
        None
    )
    
    pd.testing.assert_series_equal(results[0], expected[0])
    assert results[1] == expected[1]
    assert results[2] == expected[2]


def test_create_predict_inputs_output_ForecasterSarimax_with_last_window_and_exog():
    """
    Test _create_predict_inputs output of ForecasterSarimax with exogenous 
    variables and `last_window`.
    """
    forecaster = ForecasterSarimax(regressor = Sarimax())
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    results = forecaster._create_predict_inputs(
                  steps            = 5, 
                  exog             = exog_lw_predict_datetime, 
                  last_window      = y_lw_datetime, 
                  last_window_exog = exog_lw_datetime
              )

    expected = (
        pd.Series(
            np.array([0.59418877, 0.70775844, 0.71950195, 0.74432369, 0.80485511,
                      0.78854235, 0.9710894 , 0.84683354, 0.46382252, 0.48527317,
                      0.5280586 , 0.56233647, 0.5885704 , 0.66948036, 0.67799365,
                      0.76299549, 0.79972374, 0.77052192, 0.99438934, 0.80054443,
                      0.49055721, 0.52440799, 0.53664948, 0.55209054, 0.60336564,
                      0.68124538, 0.67807535, 0.79489265, 0.7846239 , 0.8130087 ,
                      0.9777323 , 0.89308148, 0.51269597, 0.65299589, 0.5739764 ,
                      0.63923842, 0.70387188, 0.77064824, 0.84618588, 0.89272889,
                      0.89789988, 0.94728069, 1.05070727, 0.96965567, 0.57329151,
                      0.61850684, 0.61899573, 0.66520922, 0.72652015, 0.85586494]),
            index = pd.date_range(start='2050', periods=50, freq='YE'),
            name = 'y'
        ),
        pd.DataFrame(
            {'exog': np.array([1.27501789, 1.31081724, 1.34284965, 1.37614005, 1.41412559,
                               1.45299631, 1.5056625 , 1.53112882, 1.47502858, 1.4111122 ,
                               1.35901545, 1.27726486, 1.22561223, 1.2667438 , 1.3052879 ,
                               1.35227527, 1.39975273, 1.43614303, 1.50112483, 1.52563498,
                               1.47114733, 1.41608418, 1.36930969, 1.28084993, 1.24141417,
                               1.27955181, 1.31028528, 1.36193391, 1.40844058, 1.4503692 ,
                               1.50966658, 1.55266781, 1.49622847, 1.46990287, 1.42209641,
                               1.35439763, 1.31655571, 1.36814617, 1.40678416, 1.47053466,
                               1.52226695, 1.57094872, 1.62696052, 1.65165448, 1.587767  ,
                               1.5318884 , 1.4662314 , 1.38913179, 1.34050469, 1.39701938])},
            index = pd.date_range(start='2050', periods=50, freq='YE')
        ),
        pd.DataFrame(
            {'exog': np.array([1.44651487, 1.48776549, 1.54580785, 1.58822301, 1.6196549])},
            index = pd.date_range(start='2100', periods=5, freq='YE')
        )
    )
    
    pd.testing.assert_series_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_frame_equal(results[2], expected[2])


def test_create_predict_inputs_output_ForecasterSarimax_with_last_window_and_exog_and_transformers():
    """
    Test _create_predict_inputs output of ForecasterSarimax with exogenous 
    variables, `last_window` and transformers.
    """
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterSarimax(
                     regressor        = Sarimax(), 
                     transformer_y    = StandardScaler(),
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_datetime, exog=df_exog_datetime)
    results = forecaster._create_predict_inputs(
                  steps            = 5, 
                  exog             = df_exog_lw_predict_datetime, 
                  last_window      = y_lw_datetime, 
                  last_window_exog = df_exog_lw_datetime
              )

    expected = (
        pd.Series(
            np.array([-0.07016482,  0.67641597,  0.75361509,  0.91678748,  1.31470706,
                       1.20747074,  2.4074929 ,  1.59066357, -0.92716245, -0.78615083,
                      -0.50488929, -0.27955449, -0.10709869,  0.42478471,  0.48074911,
                       1.03953158,  1.28097462,  1.0890086 ,  2.56066133,  1.28636965,
                      -0.75141477, -0.52888756, -0.44841483, -0.34690886, -0.0098382 ,
                       0.50212523,  0.48128619,  1.24921615,  1.18171176,  1.36830688,
                       2.45116179,  1.89468685, -0.60587967,  0.31641961, -0.20303627,
                       0.22598118,  0.65086662,  1.089839  ,  1.586406  ,  1.892369  ,
                       1.9263619 ,  2.25097994,  2.9308824 ,  2.39806789, -0.20753858,
                       0.08969656,  0.09291041,  0.39670722,  0.79975112,  1.65003391]),
            index = pd.date_range(start='2050', periods=50, freq='YE'),
            name = 'y'
        ),
        pd.DataFrame(
            data = np.array([[-0.20710692,  1.        ,  0.        ],
                             [ 0.10403839,  0.        ,  1.        ],
                             [ 0.38244384,  1.        ,  0.        ],
                             [ 0.67178294,  0.        ,  1.        ],
                             [ 1.00192923,  1.        ,  0.        ],
                             [ 1.33976895,  0.        ,  1.        ],
                             [ 1.79751016,  1.        ,  0.        ],
                             [ 2.01884731,  0.        ,  1.        ],
                             [ 1.5312595 ,  1.        ,  0.        ],
                             [ 0.97573875,  0.        ,  1.        ],
                             [ 0.52294675,  1.        ,  0.        ],
                             [-0.18757768,  0.        ,  1.        ],
                             [-0.63650967,  1.        ,  0.        ],
                             [-0.27902008,  0.        ,  1.        ],
                             [ 0.05598086,  1.        ,  0.        ],
                             [ 0.46436537,  0.        ,  1.        ],
                             [ 0.87700942,  1.        ,  0.        ],
                             [ 1.19329089,  0.        ,  1.        ],
                             [ 1.7580716 ,  1.        ,  0.        ],
                             [ 1.97109833,  0.        ,  1.        ],
                             [ 1.49752613,  1.        ,  0.        ],
                             [ 1.01895206,  0.        ,  1.        ],
                             [ 0.61241777,  1.        ,  0.        ],
                             [-0.15641852,  0.        ,  1.        ],
                             [-0.4991692 ,  1.        ,  0.        ],
                             [-0.16770096,  0.        ,  1.        ],
                             [ 0.09941493,  1.        ,  0.        ],
                             [ 0.54831216,  0.        ,  1.        ],
                             [ 0.95251872,  1.        ,  0.        ],
                             [ 1.31693577,  0.        ,  1.        ],
                             [ 1.83231109,  1.        ,  0.        ],
                             [ 2.20605059,  0.        ,  1.        ],
                             [ 1.71551554,  1.        ,  0.        ],
                             [ 1.48671007,  0.        ,  1.        ],
                             [ 1.07120656,  1.        ,  0.        ],
                             [ 0.48281158,  0.        ,  1.        ],
                             [ 0.15391354,  1.        ,  0.        ],
                             [ 0.60230519,  0.        ,  1.        ],
                             [ 0.93812216,  1.        ,  0.        ],
                             [ 1.49220119,  0.        ,  1.        ],
                             [ 1.94182554,  1.        ,  0.        ],
                             [ 2.3649367 ,  0.        ,  1.        ],
                             [ 2.85175584,  1.        ,  0.        ],
                             [ 3.06638012,  0.        ,  1.        ],
                             [ 2.51111055,  1.        ,  0.        ],
                             [ 2.0254491 ,  0.        ,  1.        ],
                             [ 1.45479998,  1.        ,  0.        ],
                             [ 0.78469893,  0.        ,  1.        ],
                             [ 0.36206293,  1.        ,  0.        ],
                             [ 0.85325287,  0.        ,  1.        ]]),
            columns = ['exog_1', 'exog_2_a', 'exog_2_b'],
            index = pd.date_range(start='2050', periods=50, freq='YE')
        ),
        pd.DataFrame(
            data = np.array([[1.28343637, 1.        , 0.        ],
                             [1.64196067, 0.        , 1.        ],
                             [2.14642815, 1.        , 0.        ],
                             [2.5150739 , 0.        , 1.        ],
                             [2.78826001, 1.        , 0.        ]]),
            columns = ['exog_1', 'exog_2_a', 'exog_2_b'],
            index = pd.date_range(start='2100', periods=5, freq='YE')
        )
    )
    
    pd.testing.assert_series_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_frame_equal(results[2], expected[2])


# Unit test plot history method with PyTorch backend
# ==============================================================================
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
os.environ["KERAS_BACKEND"] = "torch"
import keras
from skforecast.deep_learning.utils import create_and_compile_model
from skforecast.deep_learning import ForecasterRnn

series = pd.DataFrame(
    {
        "1": pd.Series(np.arange(50)),
        "2": pd.Series(np.arange(50)),
        "3": pd.Series(np.arange(50)),
    }
)

model = create_and_compile_model(
            series=series, 
            levels=["1", "2"],    
            lags=3,           
            steps=4,              
            recurrent_layer="LSTM",
            recurrent_units=128,
            dense_units=64,
        )
forecaster = ForecasterRnn(model, levels=["1", "2"], lags=3)
forecaster.fit(series)


def test_plot_history_with_val_loss():
    """
    Test case for the plot_history method
    """
    # Call the plot_history method
    fig, ax = plt.subplots()
    forecaster.plot_history(ax=ax)

    # Assert that the figure is of type matplotlib.figure.Figure
    assert isinstance(fig, plt.Figure)

    # Assert that the plot contains the training loss curve
    assert len(fig.axes[0].lines) == 1
    assert fig.axes[0].lines[0].get_label() == "Training Loss"

    # Assert that the plot contains the validation loss curve
    assert len(fig.axes[0].lines) == 1
    assert fig.axes[0].lines[0].get_label() == "Training Loss"
